from .distribution import (
    calculate_kl_divergence,
    evaluate_gaussian_and_tanh_log_prob,
    gaussian_and_tanh_log_prob,
    gaussian_log_prob,
    reparameterize_gaussian,
    reparameterize_gaussian_and_tanh,
)
from .input import fake_action, fake_state
from .loss import huber, quantile_loss
from .optim import clip_gradient, clip_gradient_norm, optimize, soft_update, weight_decay
from .preprocess import add_noise, get_q_at_action, get_quantile_at_action, preprocess_state
from .saving import load_params, save_params



from functools import partial

import jax
import jax.numpy as jnp


@jax.jit
def huber(td: jnp.ndarray) -> jnp.ndarray:
    """ Huber function. """
    abs_td = jnp.abs(td)
    return jnp.where(abs_td <= 1.0, jnp.square(td), abs_td)


@partial(jax.jit, static_argnums=3)
def quantile_loss(
    td: jnp.ndarray,
    cum_p: jnp.ndarray,
    weight: jnp.ndarray,
    loss_type: str,
) -> jnp.ndarray:
    """
    Calculate quantile loss.
    """
    if loss_type == "l2":
        element_wise_loss = jnp.square(td)
    elif loss_type == "huber":
        element_wise_loss = huber(td)
    else:
        NotImplementedError
    element_wise_loss *= jax.lax.stop_gradient(jnp.abs(cum_p[..., None] - (td < 0)))
    batch_loss = element_wise_loss.sum(axis=1).mean(axis=1, keepdims=True)
    return (batch_loss * weight).mean()



import numpy as np
from gym.spaces import Box


def fake_state(state_space):
    state = state_space.sample()[None, ...]
    if len(state_space.shape) == 1:
        state = state.astype(np.float32)
    return state


def fake_action(action_space):
    if type(action_space) == Box:
        action = action_space.sample().astype(np.float32)[None, ...]
    else:
        NotImplementedError
    return action



import os
from abc import abstractmethod

import numpy as np

from rljax.algorithm.base_class.base_algoirithm import OffPolicyAlgorithm
from rljax.util import fake_state, load_params, save_params


class QLearning(OffPolicyAlgorithm):
    """
    Base class for discrete Q-learning algorithms.
    """

    def __init__(
        self,
        num_agent_steps,
        state_space,
        action_space,
        seed,
        max_grad_norm,
        gamma,
        nstep,
        buffer_size,
        use_per,
        batch_size,
        start_steps,
        update_interval,
        update_interval_target,
        eps,
        eps_eval,
        eps_decay_steps,
        loss_type,
        dueling_net,
        double_q,
    ):
        assert loss_type in ["huber", "l2"]
        super(QLearning, self).__init__(
            num_agent_steps=num_agent_steps,
            state_space=state_space,
            action_space=action_space,
            seed=seed,
            max_grad_norm=max_grad_norm,
            gamma=gamma,
            nstep=nstep,
            buffer_size=buffer_size,
            use_per=use_per,
            batch_size=batch_size,
            start_steps=start_steps,
            update_interval=update_interval,
            update_interval_target=update_interval_target,
        )
        self.eps = eps
        self.eps_eval = eps_eval
        self.eps_decay_steps = eps_decay_steps
        self.loss_type = loss_type
        self.dueling_net = dueling_net
        self.double_q = double_q
        # Define fake input.
        if not hasattr(self, "fake_args"):
            self.fake_args = (fake_state(state_space),)
        # If _forward() method uses random key or not.
        if not hasattr(self, "use_key_forward"):
            self.use_key_forward = False
        # Number of random keys for _loss() method.
        if not hasattr(self, "num_keys_loss"):
            self.num_keys_loss = 0

    def select_action(self, state):
        if np.random.rand() < self.eps_eval:
            action = self.action_space.sample()
        else:
            action = self.forward(state[None, ...])
            action = np.array(action[0])
        return action

    def explore(self, state):
        if np.random.rand() < self.eps_train:
            action = self.action_space.sample()
        else:
            action = self.forward(state[None, ...])
            action = np.array(action[0])
        return action

    def forward(self, state):
        return self._forward(self.params, state, **self.kwargs_forward)

    @abstractmethod
    def _forward(self, params, state, *args, **kwargs):
        pass

    @abstractmethod
    def _calculate_value(self, params, state, action, *args, **kwargs):
        pass

    @abstractmethod
    def _calculate_target(self, params, params_target, reward, done, next_state, *args, **kwargs):
        pass

    @abstractmethod
    def _calculate_loss_and_abs_td(self, value, target, weight, *args, **kwargs):
        pass

    @property
    def eps_train(self):
        if self.agent_step > self.eps_decay_steps:
            return self.eps
        return 1.0 + (self.eps - 1.0) / self.eps_decay_steps * self.agent_step

    @property
    def kwargs_forward(self):
        return {"key": next(self.rng)} if self.use_key_forward else {}

    @property
    def kwargs_update(self):
        return {"key_list": self.get_key_list(self.num_keys_loss)} if self.num_keys_loss else {}

    def save_params(self, save_dir):
        save_params(self.params, os.path.join(save_dir, "params.npz"))

    def load_params(self, save_dir):
        self.params = self.params_target = load_params(os.path.join(save_dir, "params.npz"))



from .actor import CategoricalPolicy, DeterministicPolicy, StateDependentGaussianPolicy, StateIndependentGaussianPolicy
from .base import MLP
from .conv import DQNBody, SACDecoder, SACEncoder, SLACDecoder, SLACEncoder
from .critic import (
    ContinuousQFunction,
    ContinuousQuantileFunction,
    ContinuousVFunction,
    DiscreteImplicitQuantileFunction,
    DiscreteQFunction,
    DiscreteQuantileFunction,
)
from .misc import (
    ConstantGaussian,
    CumProbNetwork,
    Gaussian,
    SACLinear,
    make_quantile_nerwork,
    make_stochastic_latent_variable_model,
)



import os
from abc import abstractmethod
from functools import partial
from typing import List

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np

from rljax.algorithm.base_class.base_algoirithm import OffPolicyAlgorithm, OnPolicyAlgorithm
from rljax.util import fake_action, fake_state, load_params, save_params


class ActorCriticMixIn:
    """ MixIn for Actor-Critic algorithms. """

    def __init__(self):
        # If _loss_critic() method uses random key or not.
        if not hasattr(self, "use_key_critic"):
            self.use_key_critic = False
        # If _loss_actor() method uses random key or not.
        if not hasattr(self, "use_key_actor"):
            self.use_key_actor = False

    @property
    def kwargs_critic(self):
        return {"key": next(self.rng)} if self.use_key_critic else {}

    @property
    def kwargs_actor(self):
        return {"key": next(self.rng)} if self.use_key_actor else {}

    def select_action(self, state):
        action = self._select_action(self.params_actor, state[None, ...])
        return np.array(action[0])

    @abstractmethod
    def _select_action(self, params_actor, state):
        pass

    @abstractmethod
    def _explore(self, params_actor, state, key):
        pass

    def save_params(self, save_dir):
        save_params(self.params_critic, os.path.join(save_dir, "params_critic.npz"))
        save_params(self.params_actor, os.path.join(save_dir, "params_actor.npz"))

    def load_params(self, save_dir):
        self.params_critic = load_params(os.path.join(save_dir, "params_critic.npz"))
        self.params_actor = load_params(os.path.join(save_dir, "params_actor.npz"))


class OnPolicyActorCritic(ActorCriticMixIn, OnPolicyAlgorithm):
    """
    Base class for on-policy Actor-Critic algorithms.
    """

    def __init__(
        self,
        num_agent_steps,
        state_space,
        action_space,
        seed,
        max_grad_norm,
        gamma,
        buffer_size,
        batch_size,
    ):
        ActorCriticMixIn.__init__(self)
        OnPolicyAlgorithm.__init__(
            self,
            num_agent_steps=num_agent_steps,
            state_space=state_space,
            action_space=action_space,
            seed=seed,
            max_grad_norm=max_grad_norm,
            gamma=gamma,
            buffer_size=buffer_size,
            batch_size=batch_size,
        )
        # Define fake input for critic.
        if not hasattr(self, "fake_args_critic"):
            self.fake_args_critic = (fake_state(state_space),)
        # Define fake input for actor.
        if not hasattr(self, "fake_args_actor"):
            self.fake_args_actor = (fake_state(state_space),)

    def is_update(self):
        return self.agent_step % self.buffer_size == 0

    def explore(self, state):
        action, log_pi = self._explore(self.params_actor, state[None, ...], next(self.rng))
        return np.array(action[0]), np.array(log_pi[0])


class OffPolicyActorCritic(ActorCriticMixIn, OffPolicyAlgorithm):
    """
    Base class for off-policy Actor-Critic algorithms.
    """

    def __init__(
        self,
        num_agent_steps,
        state_space,
        action_space,
        seed,
        max_grad_norm,
        gamma,
        nstep,
        num_critics,
        buffer_size,
        use_per,
        batch_size,
        start_steps,
        update_interval,
        update_interval_target=None,
        tau=None,
    ):
        ActorCriticMixIn.__init__(self)
        OffPolicyAlgorithm.__init__(
            self,
            num_agent_steps=num_agent_steps,
            state_space=state_space,
            action_space=action_space,
            seed=seed,
            max_grad_norm=max_grad_norm,
            gamma=gamma,
            nstep=nstep,
            buffer_size=buffer_size,
            use_per=use_per,
            batch_size=batch_size,
            start_steps=start_steps,
            update_interval=update_interval,
            update_interval_target=update_interval_target,
            tau=tau,
        )
        self.num_critics = num_critics
        # Define fake input for critic.
        if not hasattr(self, "fake_args_critic"):
            if self.discrete_action:
                self.fake_args_critic = (fake_state(state_space),)
            else:
                self.fake_args_critic = (fake_state(state_space), fake_action(action_space))
        # Define fake input for actor.
        if not hasattr(self, "fake_args_actor"):
            self.fake_args_actor = (fake_state(state_space),)

    def explore(self, state):
        action = self._explore(self.params_actor, state[None, ...], next(self.rng))
        return np.array(action[0])

    @abstractmethod
    def _sample_action(self, params_actor, state, *args, **kwargs):
        pass

    @partial(jax.jit, static_argnums=0)
    def _calculate_value_list(
        self,
        params_critic: hk.Params,
        state: np.ndarray,
        action: np.ndarray,
    ) -> List[jnp.ndarray]:
        return self.critic.apply(params_critic, state, action)

    @partial(jax.jit, static_argnums=0)
    def _calculate_value(
        self,
        params_critic: hk.Params,
        state: np.ndarray,
        action: np.ndarray,
    ) -> jnp.ndarray:
        return jnp.asarray(self._calculate_value_list(params_critic, state, action)).min(axis=0)

    @abstractmethod
    def _calculate_target(self, params_critic_target, reward, done, next_state, next_action, *args, **kwargs):
        pass

    @partial(jax.jit, static_argnums=0)
    def _calculate_loss_critic_and_abs_td(
        self,
        value_list: List[jnp.ndarray],
        target: jnp.ndarray,
        weight: np.ndarray,
    ) -> jnp.ndarray:
        abs_td = jnp.abs(target - value_list[0])
        loss_critic = (jnp.square(abs_td) * weight).mean()
        for value in value_list[1:]:
            loss_critic += (jnp.square(target - value) * weight).mean()
        return loss_critic, jax.lax.stop_gradient(abs_td)



import numpy as np
from gym.spaces import Box, Discrete


class RolloutBuffer:
    """
    Rollout Buffer.
    """

    def __init__(
        self,
        buffer_size,
        state_space,
        action_space,
    ):
        self._n = 0
        self._p = 0
        self.buffer_size = buffer_size

        self.state = np.empty((buffer_size, *state_space.shape), dtype=np.float32)
        self.reward = np.empty((buffer_size, 1), dtype=np.float32)
        self.done = np.empty((buffer_size, 1), dtype=np.float32)
        self.log_pi = np.empty((buffer_size, 1), dtype=np.float32)
        self.next_state = np.empty((buffer_size, *state_space.shape), dtype=np.float32)

        if type(action_space) == Box:
            self.action = np.empty((buffer_size, *action_space.shape), dtype=np.float32)
        elif type(action_space) == Discrete:
            self.action = np.empty((buffer_size, 1), dtype=np.int32)
        else:
            NotImplementedError

    def append(self, state, action, reward, done, log_pi, next_state):
        self.state[self._p] = state
        self.action[self._p] = action
        self.reward[self._p] = float(reward)
        self.done[self._p] = float(done)
        self.log_pi[self._p] = float(log_pi)
        self.next_state[self._p] = next_state

        self._p = (self._p + 1) % self.buffer_size
        self._n = min(self._n + 1, self.buffer_size)

    def get(self):
        return (
            self.state,
            self.action,
            self.reward,
            self.done,
            self.log_pi,
            self.next_state,
        )



import math

import haiku as hk
import jax.numpy as jnp
import numpy as np
from jax import nn

from rljax.network.base import MLP
from rljax.network.conv import DQNBody


class ContinuousVFunction(hk.Module):
    """
    Critic for PPO.
    """

    def __init__(
        self,
        num_critics=1,
        hidden_units=(64, 64),
    ):
        super(ContinuousVFunction, self).__init__()
        self.num_critics = num_critics
        self.hidden_units = hidden_units

    def __call__(self, x):
        def _fn(x):
            return MLP(
                1,
                self.hidden_units,
                hidden_activation=jnp.tanh,
            )(x)

        if self.num_critics == 1:
            return _fn(x)
        return [_fn(x) for _ in range(self.num_critics)]


class ContinuousQFunction(hk.Module):
    """
    Critic for DDPG, TD3 and SAC.
    """

    def __init__(
        self,
        num_critics=2,
        hidden_units=(256, 256),
        d2rl=False,
    ):
        super(ContinuousQFunction, self).__init__()
        self.num_critics = num_critics
        self.hidden_units = hidden_units
        self.d2rl = d2rl

    def __call__(self, s, a):
        def _fn(x):
            return MLP(
                1,
                self.hidden_units,
                hidden_activation=nn.relu,
                hidden_scale=np.sqrt(2),
                d2rl=self.d2rl,
            )(x)

        x = jnp.concatenate([s, a], axis=1)
        # Return list even if num_critics == 1 for simple implementation.
        return [_fn(x) for _ in range(self.num_critics)]


class ContinuousQuantileFunction(hk.Module):
    """
    Critic for TQC.
    """

    def __init__(
        self,
        num_critics=5,
        hidden_units=(512, 512, 512),
        num_quantiles=25,
        d2rl=False,
    ):
        super(ContinuousQuantileFunction, self).__init__()
        self.num_critics = num_critics
        self.hidden_units = hidden_units
        self.num_quantiles = num_quantiles
        self.d2rl = d2rl

    def __call__(self, s, a):
        def _fn(x):
            return MLP(
                self.num_quantiles,
                self.hidden_units,
                hidden_activation=nn.relu,
                hidden_scale=np.sqrt(2),
                d2rl=self.d2rl,
            )(x)

        x = jnp.concatenate([s, a], axis=1)
        return [_fn(x) for _ in range(self.num_critics)]


class DiscreteQFunction(hk.Module):
    """
    Critic for DQN and SAC-Discrete.
    """

    def __init__(
        self,
        action_space,
        num_critics=1,
        hidden_units=(512,),
        dueling_net=False,
    ):
        super(DiscreteQFunction, self).__init__()
        self.action_space = action_space
        self.num_critics = num_critics
        self.hidden_units = hidden_units
        self.dueling_net = dueling_net

    def __call__(self, x):
        def _fn(x):
            if len(x.shape) == 4:
                x = DQNBody()(x)
            output = MLP(
                self.action_space.n,
                self.hidden_units,
                hidden_activation=nn.relu,
                hidden_scale=np.sqrt(2),
            )(x)
            if self.dueling_net:
                baseline = MLP(
                    1,
                    self.hidden_units,
                    hidden_activation=nn.relu,
                    hidden_scale=np.sqrt(2),
                )(x)
                return output + baseline - output.mean(axis=1, keepdims=True)
            else:
                return output

        if self.num_critics == 1:
            return _fn(x)
        return [_fn(x) for _ in range(self.num_critics)]


class DiscreteQuantileFunction(hk.Module):
    """
    Critic for QR-DQN.
    """

    def __init__(
        self,
        action_space,
        num_critics=1,
        num_quantiles=200,
        hidden_units=(512,),
        dueling_net=True,
    ):
        super(DiscreteQuantileFunction, self).__init__()
        self.action_space = action_space
        self.num_critics = num_critics
        self.num_quantiles = num_quantiles
        self.hidden_units = hidden_units
        self.dueling_net = dueling_net

    def __call__(self, x):
        def _fn(x):
            if len(x.shape) == 4:
                x = DQNBody()(x)
            output = MLP(
                self.action_space.n * self.num_quantiles,
                self.hidden_units,
                hidden_activation=nn.relu,
                hidden_scale=np.sqrt(2),
            )(x)
            output = output.reshape(-1, self.num_quantiles, self.action_space.n)
            if self.dueling_net:
                baseline = MLP(
                    self.num_quantiles,
                    self.hidden_units,
                    hidden_activation=nn.relu,
                    hidden_scale=np.sqrt(2),
                )(x)
                baseline = baseline.reshape(-1, self.num_quantiles, 1)
                return output + baseline - output.mean(axis=2, keepdims=True)
            else:
                return output

        if self.num_critics == 1:
            return _fn(x)
        return [_fn(x) for _ in range(self.num_critics)]


class DiscreteImplicitQuantileFunction(hk.Module):
    """
    Critic for IQN and FQF.
    """

    def __init__(
        self,
        action_space,
        num_critics=1,
        num_cosines=64,
        hidden_units=(512,),
        dueling_net=True,
    ):
        super(DiscreteImplicitQuantileFunction, self).__init__()
        self.action_space = action_space
        self.num_critics = num_critics
        self.num_cosines = num_cosines
        self.hidden_units = hidden_units
        self.dueling_net = dueling_net
        self.pi = math.pi * jnp.arange(1, num_cosines + 1, dtype=jnp.float32).reshape(1, 1, num_cosines)

    def __call__(self, x, cum_p):
        def _fn(x, cum_p):
            if len(x.shape) == 4:
                x = DQNBody()(x)

            # NOTE: For IQN and FQF, number of quantiles are variable.
            feature_dim = x.shape[1]
            num_quantiles = cum_p.shape[1]
            # Calculate features.
            cosine = jnp.cos(jnp.expand_dims(cum_p, 2) * self.pi).reshape(-1, self.num_cosines)
            cosine_feature = nn.relu(hk.Linear(feature_dim)(cosine)).reshape(-1, num_quantiles, feature_dim)
            x = (x.reshape(-1, 1, feature_dim) * cosine_feature).reshape(-1, feature_dim)
            # Apply quantile network.
            output = MLP(
                self.action_space.n,
                self.hidden_units,
                hidden_activation=nn.relu,
                hidden_scale=np.sqrt(2),
            )(x)
            output = output.reshape(-1, num_quantiles, self.action_space.n)
            if self.dueling_net:
                baseline = MLP(
                    1,
                    self.hidden_units,
                    hidden_activation=nn.relu,
                    hidden_scale=np.sqrt(2),
                )(x)
                baseline = baseline.reshape(-1, num_quantiles, 1)
                return output + baseline - output.mean(axis=2, keepdims=True)
            else:
                return output

        if self.num_critics == 1:
            return _fn(x, cum_p)
        return [_fn(x, cum_p) for _ in range(self.num_critics)]



import os
from time import sleep, time

import pandas as pd
from tqdm import tqdm

from rljax.algorithm.misc import SlacObservation
from rljax.trainer.base_trainer import Trainer


class SLACTrainer(Trainer):
    """
    Trainer for SLAC.
    """

    def __init__(
        self,
        env,
        env_test,
        algo,
        log_dir,
        seed=0,
        action_repeat=1,
        num_sequences=8,
        num_agent_steps=10 ** 6,
        eval_interval=10 ** 4,
        num_eval_episodes=10,
        save_params=False,
    ):
        super(SLACTrainer, self).__init__(
            env=env,
            env_test=env_test,
            algo=algo,
            log_dir=log_dir,
            seed=seed,
            action_repeat=action_repeat,
            num_agent_steps=num_agent_steps,
            eval_interval=eval_interval,
            num_eval_episodes=num_eval_episodes,
            save_params=save_params,
        )

        # Observations for training and evaluation.
        self.ob = SlacObservation(env.observation_space, env.action_space, num_sequences)
        self.ob_test = SlacObservation(env.observation_space, env.action_space, num_sequences)

    def train(self):
        # Time to start training.
        self.start_time = time()
        # Initialize the environment.
        state = self.env.reset()
        self.ob.reset_episode(state)
        self.algo.buffer.reset_episode(state)

        # Collect trajectories using random policy.
        for step in range(1, self.algo.start_steps + 1):
            self.algo.step(self.env, self.ob)

        # Update latent variable model first so that SLAC can learn well using (learned) latent dynamics.
        bar = tqdm(range(self.algo.initial_learning_steps))
        for _ in bar:
            bar.set_description("Updating latent variable model.")
            self.algo.update_model(self.writer)

        for step in range(1, self.num_agent_steps + 1):
            self.algo.step(self.env, self.ob)

            if self.algo.is_update():
                self.algo.update_model(self.writer)
                self.algo.update_sac(self.writer)

            if step % self.eval_interval == 0:
                self.evaluate(step)
                if self.save_params:
                    self.algo.save_params(os.path.join(self.param_dir, f"step{step}"))

        # Wait for the logging to be finished.
        sleep(2)

    def evaluate(self, step):
        total_return = 0.0

        for _ in range(self.num_eval_episodes):
            state = self.env_test.reset()
            self.ob_test.reset_episode(state)
            done = False
            while not done:
                action = self.algo.select_action(self.ob_test)
                state, reward, done, _ = self.env_test.step(action)
                self.ob_test.append(state, action)
                total_return += reward

        # Log mean return.
        mean_return = total_return / self.num_eval_episodes
        # To TensorBoard.
        self.writer.add_scalar("return/test", mean_return, step * self.action_repeat)
        # To CSV.
        self.log["step"].append(step * self.action_repeat)
        self.log["return"].append(mean_return)
        pd.DataFrame(self.log).to_csv(self.csv_path, index=False)

        # Log to standard output.
        print(f"Num steps: {step * self.action_repeat:<6}   Return: {mean_return:<5.1f}   Time: {self.time}")



import os
from functools import partial
from typing import List, Tuple

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
from jax.experimental import optix

from rljax.algorithm.sac import SAC
from rljax.network import ContinuousQFunction, SACDecoder, SACEncoder, SACLinear, StateDependentGaussianPolicy
from rljax.util import fake_action, fake_state, load_params, optimize, preprocess_state, save_params, soft_update, weight_decay


class SAC_AE(SAC):
    name = "SAC+AE"

    def __init__(
        self,
        num_agent_steps,
        state_space,
        action_space,
        seed,
        max_grad_norm=None,
        gamma=0.99,
        nstep=1,
        num_critics=2,
        buffer_size=10 ** 6,
        use_per=False,
        batch_size=128,
        start_steps=1000,
        update_interval=1,
        tau=0.01,
        tau_ae=0.05,
        fn_actor=None,
        fn_critic=None,
        lr_actor=1e-3,
        lr_critic=1e-3,
        lr_ae=1e-3,
        lr_alpha=1e-4,
        units_actor=(1024, 1024),
        units_critic=(1024, 1024),
        log_std_min=-10.0,
        log_std_max=2.0,
        d2rl=False,
        init_alpha=0.1,
        adam_b1_alpha=0.5,
        feature_dim=50,
        lambda_latent=1e-6,
        lambda_weight=1e-7,
        update_interval_actor=2,
        update_interval_ae=1,
        update_interval_target=2,
    ):
        assert len(state_space.shape) == 3 and state_space.shape[:2] == (84, 84)
        assert (state_space.high == 255).all()
        if d2rl:
            self.name += "-D2RL"

        if fn_critic is None:

            def fn_critic(x, a):
                # Define without linear layer.
                return ContinuousQFunction(
                    num_critics=num_critics,
                    hidden_units=units_critic,
                    d2rl=d2rl,
                )(x, a)

        if fn_actor is None:

            def fn_actor(x):
                # Define with linear layer.
                x = SACLinear(feature_dim=feature_dim)(x)
                return StateDependentGaussianPolicy(
                    action_space=action_space,
                    hidden_units=units_actor,
                    log_std_min=log_std_min,
                    log_std_max=log_std_max,
                    clip_log_std=False,
                    d2rl=d2rl,
                )(x)

        fake_feature = jnp.empty((1, feature_dim))
        fake_last_conv = jnp.empty((1, 39200))
        if not hasattr(self, "fake_args_critic"):
            self.fake_args_critic = (fake_feature, fake_action(action_space))
        if not hasattr(self, "fake_args_actor"):
            self.fake_args_actor = (fake_last_conv,)

        super(SAC_AE, self).__init__(
            num_agent_steps=num_agent_steps,
            state_space=state_space,
            action_space=action_space,
            seed=seed,
            max_grad_norm=max_grad_norm,
            gamma=gamma,
            nstep=nstep,
            num_critics=num_critics,
            buffer_size=buffer_size,
            use_per=use_per,
            batch_size=batch_size,
            start_steps=start_steps,
            update_interval=update_interval,
            tau=tau,
            fn_actor=fn_actor,
            fn_critic=fn_critic,
            lr_actor=lr_actor,
            lr_critic=lr_critic,
            lr_alpha=lr_alpha,
            init_alpha=init_alpha,
            adam_b1_alpha=adam_b1_alpha,
        )
        # Encoder.
        self.encoder = hk.without_apply_rng(hk.transform(lambda s: SACEncoder(num_filters=32, num_layers=4)(s)))
        self.params_encoder = self.params_encoder_target = self.encoder.init(next(self.rng), fake_state(state_space))

        # Linear layer for critic and decoder.
        self.linear = hk.without_apply_rng(hk.transform(lambda x: SACLinear(feature_dim=feature_dim)(x)))
        self.params_linear = self.params_linear_target = self.linear.init(next(self.rng), fake_last_conv)

        # Decoder.
        self.decoder = hk.without_apply_rng(hk.transform(lambda x: SACDecoder(state_space, num_filters=32, num_layers=4)(x)))
        self.params_decoder = self.decoder.init(next(self.rng), fake_feature)
        opt_init, self.opt_ae = optix.adam(lr_ae)
        self.opt_state_ae = opt_init(self.params_ae)

        # Re-define the optimizer for critic.
        opt_init, self.opt_critic = optix.adam(lr_critic)
        self.opt_state_critic = opt_init(self.params_entire_critic)

        # Other parameters.
        self._update_target_ae = jax.jit(partial(soft_update, tau=tau_ae))
        self.lambda_latent = lambda_latent
        self.lambda_weight = lambda_weight
        self.update_interval_actor = update_interval_actor
        self.update_interval_ae = update_interval_ae
        self.update_interval_target = update_interval_target

    def select_action(self, state):
        last_conv = self._preprocess(self.params_encoder, state[None, ...])
        action = self._select_action(self.params_actor, last_conv)
        return np.array(action[0])

    def explore(self, state):
        last_conv = self._preprocess(self.params_encoder, state[None, ...])
        action = self._explore(self.params_actor, last_conv, next(self.rng))
        return np.array(action[0])

    @partial(jax.jit, static_argnums=0)
    def _preprocess(
        self,
        params_encoder: hk.Params,
        state: np.ndarray,
    ) -> jnp.ndarray:
        return self.encoder.apply(params_encoder, state)

    def update(self, writer=None):
        self.learning_step += 1
        weight, batch = self.buffer.sample(self.batch_size)
        state, action, reward, done, next_state = batch

        # Update critic.
        self.opt_state_critic, params_entire_critic, loss_critic, abs_td = optimize(
            self._loss_critic,
            self.opt_critic,
            self.opt_state_critic,
            self.params_entire_critic,
            self.max_grad_norm,
            params_critic_target=self.params_entire_critic_target,
            params_actor=self.params_actor,
            log_alpha=self.log_alpha,
            state=state,
            action=action,
            reward=reward,
            done=done,
            next_state=next_state,
            weight=weight,
            **self.kwargs_critic,
        )
        self.params_encoder = params_entire_critic["encoder"]
        self.params_linear = params_entire_critic["linear"]
        self.params_critic = params_entire_critic["critic"]

        # Update priority.
        if self.use_per:
            self.buffer.update_priority(abs_td)

        # Update actor and alpha.
        if self.learning_step % self.update_interval_actor == 0:
            self.opt_state_actor, self.params_actor, loss_actor, mean_log_pi = optimize(
                self._loss_actor,
                self.opt_actor,
                self.opt_state_actor,
                self.params_actor,
                self.max_grad_norm,
                params_critic=self.params_entire_critic,
                log_alpha=self.log_alpha,
                state=state,
                **self.kwargs_actor,
            )
            self.opt_state_alpha, self.log_alpha, loss_alpha, _ = optimize(
                self._loss_alpha,
                self.opt_alpha,
                self.opt_state_alpha,
                self.log_alpha,
                None,
                mean_log_pi=mean_log_pi,
            )

        # Update autoencoder.
        if self.learning_step % self.update_interval_actor == 0:
            self.opt_state_ae, params_ae, loss_ae, _ = optimize(
                self._loss_ae,
                self.opt_ae,
                self.opt_state_ae,
                self.params_ae,
                self.max_grad_norm,
                state=state,
                key=next(self.rng),
            )
            self.params_encoder = params_ae["encoder"]
            self.params_linear = params_ae["linear"]
            self.params_decoder = params_ae["decoder"]

        # Update target network.
        if self.learning_step % self.update_interval_target == 0:
            self.params_encoder_target = self._update_target_ae(self.params_encoder_target, self.params_encoder)
            self.params_linear_target = self._update_target_ae(self.params_linear_target, self.params_linear)
            self.params_critic_target = self._update_target(self.params_critic_target, self.params_critic)

        if writer and self.learning_step % 1000 == 0:
            writer.add_scalar("loss/critic", loss_critic, self.learning_step)
            writer.add_scalar("loss/actor", loss_actor, self.learning_step)
            writer.add_scalar("loss/ae", loss_ae, self.learning_step)
            writer.add_scalar("loss/alpha", loss_alpha, self.learning_step)
            writer.add_scalar("stat/alpha", jnp.exp(self.log_alpha), self.learning_step)
            writer.add_scalar("stat/entropy", -mean_log_pi, self.learning_step)

    @partial(jax.jit, static_argnums=0)
    def _calculate_value_list(
        self,
        params_critic: hk.Params,
        last_conv: np.ndarray,
        action: np.ndarray,
    ) -> List[jnp.ndarray]:
        feature = self.linear.apply(params_critic["linear"], last_conv)
        return self.critic.apply(params_critic["critic"], feature, action)

    @partial(jax.jit, static_argnums=0)
    def _loss_critic(
        self,
        params_critic: hk.Params,
        params_critic_target: hk.Params,
        params_actor: hk.Params,
        log_alpha: jnp.ndarray,
        state: np.ndarray,
        action: np.ndarray,
        reward: np.ndarray,
        done: np.ndarray,
        next_state: np.ndarray,
        weight: np.ndarray or List[jnp.ndarray],
        *args,
        **kwargs,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        last_conv = self.encoder.apply(params_critic["encoder"], state)
        next_last_conv = jax.lax.stop_gradient(self.encoder.apply(params_critic["encoder"], next_state))
        return super(SAC_AE, self)._loss_critic(
            params_critic=params_critic,
            params_critic_target=params_critic_target,
            params_actor=params_actor,
            log_alpha=log_alpha,
            state=last_conv,
            action=action,
            reward=reward,
            done=done,
            next_state=next_last_conv,
            weight=weight,
            *args,
            **kwargs,
        )

    @partial(jax.jit, static_argnums=0)
    def _loss_actor(
        self,
        params_actor: hk.Params,
        params_critic: hk.Params,
        log_alpha: jnp.ndarray,
        state: np.ndarray,
        *args,
        **kwargs,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        last_conv = jax.lax.stop_gradient(self.encoder.apply(params_critic["encoder"], state))
        return super(SAC_AE, self)._loss_actor(
            params_actor=params_actor,
            params_critic=params_critic,
            log_alpha=log_alpha,
            state=last_conv,
            *args,
            **kwargs,
        )

    @partial(jax.jit, static_argnums=0)
    def _loss_ae(
        self,
        params_ae: hk.Params,
        state: np.ndarray,
        key: jnp.ndarray,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        # Preprocess states.
        target = preprocess_state(state, key)
        # Reconstruct states.
        last_conv = self.encoder.apply(params_ae["encoder"], state)
        feature = self.linear.apply(params_ae["linear"], last_conv)
        reconst = self.decoder.apply(params_ae["decoder"], feature)
        # MSE for reconstruction errors.
        loss_reconst = jnp.square(target - reconst).mean()
        # L2 penalty of latent representations following RAE.
        loss_latent = 0.5 * jnp.square(feature).sum(axis=1).mean()
        # Weight decay for the decoder.
        loss_weight = weight_decay(params_ae["decoder"])
        return loss_reconst + self.lambda_latent * loss_latent + self.lambda_weight * loss_weight, None

    @property
    def params_ae(self):
        return {
            "encoder": self.params_encoder,
            "linear": self.params_linear,
            "decoder": self.params_decoder,
        }

    @property
    def params_entire_critic(self):
        return {
            "encoder": self.params_encoder,
            "linear": self.params_linear,
            "critic": self.params_critic,
        }

    @property
    def params_entire_critic_target(self):
        return {
            "encoder": self.params_encoder_target,
            "linear": self.params_linear_target,
            "critic": self.params_critic_target,
        }

    def save_params(self, save_dir):
        super().save_params(save_dir)
        save_params(self.params_encoder, os.path.join(save_dir, "params_encoder.npz"))
        save_params(self.params_linear, os.path.join(save_dir, "params_linear.npz"))
        save_params(self.params_decoder, os.path.join(save_dir, "params_decoder.npz"))

    def load_params(self, save_dir):
        super().load_params(save_dir)
        self.params_encoder = self.params_encoder_target = load_params(os.path.join(save_dir, "params_encoder.npz"))
        self.params_linear = self.params_linear_target = load_params(os.path.join(save_dir, "params_linear.npz"))
        self.params_decoder = load_params(os.path.join(save_dir, "params_decoder.npz"))



import haiku as hk
import jax.numpy as jnp
from jax import nn


class MLP(hk.Module):
    def __init__(
        self,
        output_dim,
        hidden_units,
        hidden_activation=nn.relu,
        output_activation=None,
        hidden_scale=1.0,
        output_scale=1.0,
        d2rl=False,
    ):
        super(MLP, self).__init__()
        self.output_dim = output_dim
        self.hidden_units = hidden_units
        self.hidden_activation = hidden_activation
        self.output_activation = output_activation
        self.d2rl = d2rl
        self.hidden_kwargs = {"w_init": hk.initializers.Orthogonal(scale=hidden_scale)}
        self.output_kwargs = {"w_init": hk.initializers.Orthogonal(scale=output_scale)}

    def __call__(self, x):
        x_input = x
        for i, unit in enumerate(self.hidden_units):
            x = hk.Linear(unit, **self.hidden_kwargs)(x)
            x = self.hidden_activation(x)
            if self.d2rl and i + 1 != len(self.hidden_units):
                x = jnp.concatenate([x, x_input], axis=1)
        x = hk.Linear(self.output_dim, **self.output_kwargs)(x)
        if self.output_activation is not None:
            x = self.output_activation(x)
        return x



from functools import partial
from typing import List, Tuple

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
from jax.experimental import optix

from rljax.algorithm.misc import SlacMixIn
from rljax.algorithm.sac import SAC
from rljax.network import ContinuousQFunction, StateDependentGaussianPolicy, make_stochastic_latent_variable_model
from rljax.util import calculate_kl_divergence, gaussian_log_prob, optimize


class SLAC(SlacMixIn, SAC):
    name = "SLAC"

    def __init__(
        self,
        num_agent_steps,
        state_space,
        action_space,
        seed,
        max_grad_norm=None,
        gamma=0.99,
        num_sequences=8,
        num_critics=2,
        buffer_size=50000,
        batch_size_sac=256,
        batch_size_model=32,
        start_steps=10000,
        initial_learning_steps=50,
        update_interval=1,
        tau=5e-3,
        fn_actor=None,
        fn_critic=None,
        lr_actor=3e-4,
        lr_critic=3e-4,
        lr_alpha=3e-4,
        lr_model=1e-4,
        units_actor=(256, 256),
        units_critic=(256, 256),
        units_model=(256, 256),
        log_std_min=-20.0,
        log_std_max=2.0,
        d2rl=False,
        init_alpha=1.0,
        adam_b1_alpha=0.9,
        feature_dim=256,
        z1_dim=32,
        z2_dim=256,
    ):
        assert len(state_space.shape) == 3 and state_space.shape[:2] == (64, 64)
        assert (state_space.high == 255).all()
        SlacMixIn.__init__(
            self,
            state_space=state_space,
            action_space=action_space,
            num_sequences=num_sequences,
            buffer_size=buffer_size,
            batch_size_sac=batch_size_sac,
            batch_size_model=batch_size_model,
            initial_learning_steps=initial_learning_steps,
            feature_dim=feature_dim,
            z1_dim=z1_dim,
            z2_dim=z2_dim,
        )
        if d2rl:
            self.name += "-D2RL"

        if fn_critic is None:

            def fn_critic(z, a):
                return ContinuousQFunction(
                    num_critics=num_critics,
                    hidden_units=units_critic,
                    d2rl=d2rl,
                )(z, a)

        if fn_actor is None:

            def fn_actor(x):
                return StateDependentGaussianPolicy(
                    action_space=action_space,
                    hidden_units=units_actor,
                    log_std_min=log_std_min,
                    log_std_max=log_std_max,
                    d2rl=d2rl,
                )(x)

        SAC.__init__(
            self,
            num_agent_steps=num_agent_steps,
            state_space=state_space,
            action_space=action_space,
            seed=seed,
            max_grad_norm=max_grad_norm,
            gamma=gamma,
            nstep=1,
            num_critics=num_critics,
            buffer_size=None,
            use_per=False,
            batch_size=None,
            start_steps=start_steps,
            update_interval=update_interval,
            tau=tau,
            fn_actor=fn_actor,
            fn_critic=fn_critic,
            lr_actor=lr_actor,
            lr_critic=lr_critic,
            lr_alpha=lr_alpha,
            init_alpha=init_alpha,
            adam_b1_alpha=adam_b1_alpha,
        )
        self.model, self.params_model = make_stochastic_latent_variable_model(
            rng=self.rng,
            state_space=state_space,
            action_space=action_space,
            num_sequences=num_sequences,
            units_model=units_model,
            z1_dim=z1_dim,
            z2_dim=z2_dim,
            feature_dim=feature_dim,
        )
        opt_init, self.opt_model = optix.adam(lr_model)
        self.opt_state_model = opt_init(self.params_model)

    @partial(jax.jit, static_argnums=0)
    def _preprocess(
        self,
        params_model: hk.Params,
        state: np.ndarray,
        action: np.ndarray,
    ) -> jnp.ndarray:
        feature = self.model["encoder"].apply(params_model["encoder"], state).reshape([1, -1])
        return jnp.concatenate([feature, action], axis=1)

    @partial(jax.jit, static_argnums=0)
    def get_input_for_sac(
        self,
        params_model: hk.Params,
        state_: np.ndarray,
        action_: np.ndarray,
        key_list: List[np.ndarray],
    ) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
        N = state_.shape[0]
        feature_ = self.model["encoder"].apply(params_model["encoder"], state_)
        z_ = jax.lax.stop_gradient(jnp.concatenate(self.sample_post(params_model, feature_, action_, key_list)[2:], axis=-1))
        feature_action = jnp.concatenate([feature_[:, :-1].reshape([N, -1]), action_[:, :-1].reshape([N, -1])], axis=-1)
        next_feature_action = jnp.concatenate([feature_[:, 1:].reshape([N, -1]), action_[:, 1:].reshape([N, -1])], axis=-1)
        return z_[:, -2], z_[:, -1], action_[:, -1], feature_action, next_feature_action

    def update_sac(self, writer=None):
        self.learning_step_sac += 1
        state_, action_, reward, done = self.buffer.sample_sac(self.batch_size_sac)
        z, next_z, action, feature_action, next_feature_action = self.get_input_for_sac(
            params_model=self.params_model,
            state_=state_,
            action_=action_,
            key_list=[next(self.rng) for _ in range(2 * (self.num_sequences + 1))],
        )
        del state_, action_

        # Update critic.
        self.opt_state_critic, self.params_critic, loss_critic, _ = optimize(
            self._loss_critic,
            self.opt_critic,
            self.opt_state_critic,
            self.params_critic,
            self.max_grad_norm,
            params_critic_target=self.params_critic_target,
            params_actor=self.params_actor,
            log_alpha=self.log_alpha,
            z=z,
            next_z=next_z,
            action=action,
            reward=reward,
            done=done,
            next_feature_action=next_feature_action,
            **self.kwargs_critic,
        )

        # Update actor and alpha.
        self.opt_state_actor, self.params_actor, loss_actor, mean_log_pi = optimize(
            self._loss_actor,
            self.opt_actor,
            self.opt_state_actor,
            self.params_actor,
            self.max_grad_norm,
            params_critic=self.params_critic,
            log_alpha=self.log_alpha,
            z=z,
            feature_action=feature_action,
            **self.kwargs_actor,
        )
        self.opt_state_alpha, self.log_alpha, loss_alpha, _ = optimize(
            self._loss_alpha,
            self.opt_alpha,
            self.opt_state_alpha,
            self.log_alpha,
            None,
            mean_log_pi=mean_log_pi,
        )

        # Update target network.
        self.params_critic_target = self._update_target(self.params_critic_target, self.params_critic)

        if writer and self.learning_step_sac % 1000 == 0:
            writer.add_scalar("loss/critic", loss_critic, self.learning_step_sac)
            writer.add_scalar("loss/actor", loss_actor, self.learning_step_sac)
            writer.add_scalar("loss/alpha", loss_alpha, self.learning_step_sac)
            writer.add_scalar("stat/alpha", jnp.exp(self.log_alpha), self.learning_step_sac)
            writer.add_scalar("stat/entropy", -mean_log_pi, self.learning_step_sac)

    @partial(jax.jit, static_argnums=0)
    def _calculate_target(
        self,
        params_critic_target: hk.Params,
        log_alpha: jnp.ndarray,
        reward: np.ndarray,
        done: np.ndarray,
        next_z: np.ndarray,
        next_action: jnp.ndarray,
        next_log_pi: jnp.ndarray,
    ) -> jnp.ndarray:
        alpha = jnp.exp(log_alpha)
        next_q_list = self.critic.apply(params_critic_target, next_z, next_action)
        next_q = jnp.asarray(next_q_list).min(axis=0) - alpha * next_log_pi
        return jax.lax.stop_gradient(reward + (1.0 - done) * self.discount * next_q)

    @partial(jax.jit, static_argnums=0)
    def _loss_critic(
        self,
        params_critic: hk.Params,
        params_critic_target: hk.Params,
        params_actor: hk.Params,
        log_alpha: jnp.ndarray,
        z: np.ndarray,
        next_z: np.ndarray,
        action: np.ndarray,
        reward: np.ndarray,
        done: np.ndarray,
        next_feature_action: np.ndarray,
        *args,
        **kwargs,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        next_action, next_log_pi = self._sample_action(params_actor, next_feature_action, *args, **kwargs)
        target = self._calculate_target(params_critic_target, log_alpha, reward, done, next_z, next_action, next_log_pi)
        q_list = self._calculate_value_list(params_critic, z, action)
        return self._calculate_loss_critic_and_abs_td(q_list, target, 1.0)

    @partial(jax.jit, static_argnums=0)
    def _loss_actor(
        self,
        params_actor: hk.Params,
        params_critic: hk.Params,
        log_alpha: jnp.ndarray,
        z: np.ndarray,
        feature_action: np.ndarray,
        *args,
        **kwargs,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        action, log_pi = self._sample_action(params_actor=params_actor, state=feature_action, *args, **kwargs)
        mean_q = self._calculate_value(params_critic, z, action).mean()
        mean_log_pi = self._calculate_log_pi(action, log_pi).mean()
        return jax.lax.stop_gradient(jnp.exp(log_alpha)) * mean_log_pi - mean_q, jax.lax.stop_gradient(mean_log_pi)

    def update_model(self, writer=None):
        self.learning_step_model += 1
        state_, action_, reward_, done_ = self.buffer.sample_model(self.batch_size_model)

        self.opt_state_model, self.params_model, loss_model, _ = optimize(
            self._loss_model,
            self.opt_model,
            self.opt_state_model,
            self.params_model,
            self.max_grad_norm,
            state_=state_,
            action_=action_,
            reward_=reward_,
            done_=done_,
            key_list1=self.get_key_list(2 * (self.num_sequences + 1)),
            key_list2=self.get_key_list(2 * (self.num_sequences + 1)),
        )

        if writer and self.learning_step_model % 1000 == 0:
            writer.add_scalar("loss/latent", loss_model, self.learning_step_model)

    @partial(jax.jit, static_argnums=0)
    def _loss_model(
        self,
        params_model: hk.Params,
        state_: np.ndarray,
        action_: np.ndarray,
        reward_: np.ndarray,
        done_: np.ndarray,
        key_list1: List[np.ndarray],
        key_list2: List[np.ndarray],
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        # Floatify the sequence of images.
        img_ = state_.astype(jnp.float32) / 255.0
        # Calculate the sequence of features.
        feature_ = self.model["encoder"].apply(params_model["encoder"], state_)
        # Sample from stochastic latent variable model.
        z1_mean_pri_, z1_std_pri_ = self.sample_prior(params_model, action_, key_list1)
        z1_mean_post_, z1_std_post_, z1_, z2_ = self.sample_post(params_model, feature_, action_, key_list2)

        # Calculate KL divergence loss.
        loss_kld = calculate_kl_divergence(z1_mean_post_, z1_std_post_, z1_mean_pri_, z1_std_pri_).mean(axis=0).sum()
        # Prediction loss of images.
        img_mean_, img_std_ = self.model["decoder"].apply(params_model["decoder"], z1_, z2_)
        loss_img = -gaussian_log_prob(jnp.log(img_std_), (img_ - img_mean_) / img_std_).mean(axis=0).sum()
        # Prediction loss of rewards.
        z_ = jnp.concatenate([z1_, z2_], axis=-1)
        rew_mean_, rew_std_ = self.model["reward"].apply(params_model["reward"], z_[:, :-1], action_, z_[:, 1:])
        loss_rew = -(gaussian_log_prob(jnp.log(rew_std_), (reward_ - rew_mean_) / rew_std_) * (1 - done_)).mean(axis=0).sum()
        return loss_kld + loss_img + loss_rew, None

    @partial(jax.jit, static_argnums=0)
    def sample_prior(
        self,
        params_model: hk.Params,
        action_: jnp.ndarray,
        key_list: List[np.ndarray],
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        z1_mean_ = []
        z1_std_ = []
        for t in range(self.num_sequences + 1):
            if t == 0:
                z1_mean, z1_std = self.model["z1_prior_init"].apply(params_model["z1_prior_init"], action_[:, 0])
                z1 = z1_mean + jax.random.normal(key_list[0], z1_std.shape) * z1_std
                z2_mean, z2_std = self.model["z2_init"].apply(params_model["z2_init"], z1)
                z2 = z2_mean + jax.random.normal(key_list[1], z2_std.shape) * z2_std
            else:
                z1_mean, z1_std = self.model["z1_prior"].apply(params_model["z1_prior"], z2, action_[:, t - 1])
                z1 = z1_mean + jax.random.normal(key_list[2 * t], z1_std.shape) * z1_std
                z2_mean, z2_std = self.model["z2"].apply(params_model["z2"], z1, z2, action_[:, t - 1])
                z2 = z2_mean + jax.random.normal(key_list[2 * t + 1], z2_std.shape) * z2_std
            z1_mean_.append(z1_mean)
            z1_std_.append(z1_std)
        return (jnp.stack(z1_mean_, axis=1), jnp.stack(z1_std_, axis=1))

    @partial(jax.jit, static_argnums=0)
    def sample_post(
        self,
        params_model: hk.Params,
        feature_: jnp.ndarray,
        action_: jnp.ndarray,
        key_list: List[np.ndarray],
    ) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
        z1_mean_ = []
        z1_std_ = []
        z1_ = []
        z2_ = []
        for t in range(self.num_sequences + 1):
            if t == 0:
                z1_mean, z1_std = self.model["z1_post_init"].apply(params_model["z1_post_init"], feature_[:, 0])
                z1 = z1_mean + jax.random.normal(key_list[0], z1_std.shape) * z1_std
                z2_mean, z2_std = self.model["z2_init"].apply(params_model["z2_init"], z1)
                z2 = z2_mean + jax.random.normal(key_list[1], z2_std.shape) * z2_std
            else:
                z1_mean, z1_std = self.model["z1_post"].apply(params_model["z1_post"], feature_[:, t], z2, action_[:, t - 1])
                z1 = z1_mean + jax.random.normal(key_list[2 * t], z1_std.shape) * z1_std
                z2_mean, z2_std = self.model["z2"].apply(params_model["z2"], z1, z2, action_[:, t - 1])
                z2 = z2_mean + jax.random.normal(key_list[2 * t + 1], z2_std.shape) * z2_std
            z1_mean_.append(z1_mean)
            z1_std_.append(z1_std)
            z1_.append(z1)
            z2_.append(z2)
        return (jnp.stack(z1_mean_, axis=1), jnp.stack(z1_std_, axis=1), jnp.stack(z1_, axis=1), jnp.stack(z2_, axis=1))



from functools import partial
from typing import Any, Tuple

import haiku as hk
import jax
import jax.numpy as jnp
from jax.experimental import optix
from jax.tree_util import tree_flatten


@partial(jax.jit, static_argnums=(0, 1, 4))
def optimize(
    fn_loss: Any,
    opt: Any,
    opt_state: Any,
    params_to_update: hk.Params,
    max_grad_norm: float or None,
    *args,
    **kwargs,
) -> Tuple[Any, hk.Params, jnp.ndarray, Any]:
    (loss, aux), grad = jax.value_and_grad(fn_loss, has_aux=True)(
        params_to_update,
        *args,
        **kwargs,
    )
    if max_grad_norm is not None:
        grad = clip_gradient_norm(grad, max_grad_norm)
    update, opt_state = opt(grad, opt_state)
    params_to_update = optix.apply_updates(params_to_update, update)
    return opt_state, params_to_update, loss, aux


@jax.jit
def clip_gradient(
    grad: Any,
    max_value: float,
) -> Any:
    """
    Clip gradients.
    """
    return jax.tree_map(lambda g: jnp.clip(g, -max_value, max_value), grad)


@jax.jit
def clip_gradient_norm(
    grad: Any,
    max_grad_norm: float,
) -> Any:
    """
    Clip norms of gradients.
    """

    def _clip_gradient_norm(g):
        clip_coef = max_grad_norm / (jax.lax.stop_gradient(jnp.linalg.norm(g)) + 1e-6)
        clip_coef = jnp.clip(clip_coef, a_max=1.0)
        return g * clip_coef

    return jax.tree_map(lambda g: _clip_gradient_norm(g), grad)


@jax.jit
def soft_update(
    target_params: hk.Params,
    online_params: hk.Params,
    tau: float,
) -> hk.Params:
    """
    Update target network using Polyak-Ruppert Averaging.
    """
    return jax.tree_multimap(lambda t, s: (1 - tau) * t + tau * s, target_params, online_params)


@jax.jit
def weight_decay(params: hk.Params) -> jnp.ndarray:
    """
    Calculate the sum of L2 norms of parameters.
    """
    leaves, _ = tree_flatten(params)
    return 0.5 * sum(jnp.vdot(x, x) for x in leaves)



from functools import partial

import jax
import jax.numpy as jnp
import numpy as np

from rljax.buffer.replay_buffer import ReplayBuffer
from rljax.buffer.segment_tree import MinTree, SumTree


class PrioritizedReplayBuffer(ReplayBuffer):
    """
    Prioritized Replay Buffer.
    """

    def __init__(
        self,
        buffer_size,
        state_space,
        action_space,
        gamma,
        nstep,
        alpha=0.6,
        beta=0.4,
        beta_steps=10 ** 5,
        min_pa=0.0,
        max_pa=1.0,
        eps=0.01,
    ):
        super(PrioritizedReplayBuffer, self).__init__(
            buffer_size=buffer_size,
            state_space=state_space,
            action_space=action_space,
            gamma=gamma,
            nstep=nstep,
        )

        self.alpha = alpha
        self.beta = beta
        self.beta_diff = (1.0 - beta) / beta_steps
        self.min_pa = min_pa
        self.max_pa = max_pa
        self.eps = eps
        self._cached_idxes = None

        tree_size = 1
        while tree_size < buffer_size:
            tree_size *= 2
        self.tree_sum = SumTree(tree_size)
        self.tree_min = MinTree(tree_size)

    def _append(self, state, action, reward, next_state, done):
        # Assign max priority when stored for the first time.
        self.tree_min[self._p] = self.max_pa
        self.tree_sum[self._p] = self.max_pa
        super()._append(state, action, reward, next_state, done)

    def _sample_idx(self, batch_size):
        total_pa = self.tree_sum.reduce(0, self._n)
        rand = np.random.rand(batch_size) * total_pa
        idxes = [self.tree_sum.find_prefixsum_idx(r) for r in rand]
        self.beta = min(1.0, self.beta + self.beta_diff)
        return idxes

    def sample(self, batch_size):
        assert self._cached_idxes is None, "Update priorities before sampling."

        self._cached_idxes = self._sample_idx(batch_size)
        weight = self._calculate_weight(self._cached_idxes)
        batch = self._sample(self._cached_idxes)
        return weight, batch

    def _calculate_weight(self, idxes):
        min_pa = self.tree_min.reduce(0, self._n)
        weight = [(self.tree_sum[i] / min_pa) ** -self.beta for i in idxes]
        weight = np.array(weight, dtype=np.float32)
        return np.expand_dims(weight, axis=1)

    def update_priority(self, abs_td):
        assert self._cached_idxes is not None, "Sample batch before updating priorities."
        assert abs_td.shape[1:] == (1,)
        pa = np.array(self._calculate_pa(abs_td), dtype=np.float32).flatten()
        for i, idx in enumerate(self._cached_idxes):
            self.tree_sum[idx] = pa[i]
            self.tree_min[idx] = pa[i]
        self._cached_idxes = None

    @partial(jax.jit, static_argnums=0)
    def _calculate_pa(self, abs_td: jnp.ndarray) -> jnp.ndarray:
        return jnp.clip((abs_td + self.eps) ** self.alpha, self.min_pa, self.max_pa)



import operator


class SegmentTree:
    """
    Segment tree.
    """

    def __init__(
        self,
        size,
        op,
        value_init,
    ):
        assert size > 0 and size & (size - 1) == 0
        self._size = size
        self._op = op
        self._value_init = value_init
        self._values = [value_init for _ in range(2 * size)]

    def reduce(self, start, end):
        start += self._size
        end += self._size

        res = self._value_init
        while start < end:
            if start & 1:
                res = self._op(res, self._values[start])
                start += 1

            if end & 1:
                end -= 1
                res = self._op(res, self._values[end])

            start //= 2
            end //= 2

        return res

    def __setitem__(self, idx, val):
        assert 0 <= idx < self._size

        # Set value.
        idx += self._size
        self._values[idx] = val

        # Update its ancestors iteratively.
        idx = idx >> 1
        while idx >= 1:
            left = 2 * idx
            self._values[idx] = self._op(self._values[left], self._values[left + 1])
            idx = idx >> 1

    def __getitem__(self, idx):
        assert 0 <= idx < self._size
        return self._values[idx + self._size]


class SumTree(SegmentTree):
    """
    Sum tree.
    """

    def __init__(self, size):
        super().__init__(size, operator.add, 0.0)

    def find_prefixsum_idx(self, prefixsum):
        # Traverse to the leaf.
        idx = 1
        while idx < self._size:
            left = 2 * idx
            if self._values[left] > prefixsum:
                idx = left
            else:
                prefixsum -= self._values[left]
                idx = left + 1
        return idx - self._size


class MinTree(SegmentTree):
    """
    Min tree.
    """

    def __init__(self, size):
        super().__init__(size, min, float("inf"))



import os

import haiku as hk
import numpy as np


def save_params(params, path):
    """
    Save parameters.
    """
    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
    np.savez(path, **params)


def load_params(path):
    """
    Load parameters.
    """
    return hk.data_structures.to_immutable_dict(np.load(path))



import os
from abc import abstractmethod
from collections import deque

import jax.numpy as jnp
import numpy as np

from rljax.buffer import SLACReplayBuffer
from rljax.util import fake_action, load_params, save_params


class SlacObservation:
    """
    Observation for SLAC.
    """

    def __init__(self, state_space, action_space, num_sequences):
        self.state_shape = state_space.shape
        self.action_shape = action_space.shape
        self.num_sequences = num_sequences

    def reset_episode(self, state):
        self._state = deque(maxlen=self.num_sequences)
        self._action = deque(maxlen=self.num_sequences - 1)
        for _ in range(self.num_sequences - 1):
            self._state.append(np.zeros(self.state_shape, dtype=np.uint8))
            self._action.append(np.zeros(self.action_shape, dtype=np.float32))
        self._state.append(state)

    def append(self, state, action):
        self._state.append(state)
        self._action.append(action)

    @property
    def state(self):
        return np.array(self._state, dtype=np.uint8)[None, ...]

    @property
    def action(self):
        return np.array(self._action, dtype=np.float32).reshape(1, -1)


class SlacMixIn:
    """
    MixIn for SLAC-based algorithms.
    """

    def __init__(
        self,
        state_space,
        action_space,
        num_sequences,
        buffer_size,
        batch_size_sac,
        batch_size_model,
        initial_learning_steps,
        feature_dim,
        z1_dim,
        z2_dim,
    ):
        self.buffer = SLACReplayBuffer(
            buffer_size=buffer_size,
            state_space=state_space,
            action_space=action_space,
            num_sequences=num_sequences,
        )
        # Define fake input for critic.
        if not hasattr(self, "fake_args_critic"):
            fake_z = jnp.empty((1, z1_dim + z2_dim))
            self.fake_args_critic = (fake_z, fake_action(action_space))
        # Define fake input for actor.
        if not hasattr(self, "fake_args_actor"):
            fake_feature_action = jnp.empty((1, num_sequences * feature_dim + (num_sequences - 1) * action_space.shape[0]))
            self.fake_args_actor = (fake_feature_action,)

        self.learning_step_model = 0
        self.learning_step_sac = 0
        self.num_sequences = num_sequences
        self.batch_size_sac = batch_size_sac
        self.batch_size_model = batch_size_model
        self.initial_learning_steps = initial_learning_steps

    def step(self, env, ob):
        self.agent_step += 1
        self.episode_step += 1

        if self.agent_step <= self.start_steps:
            action = env.action_space.sample()
        else:
            action = self.explore(ob)

        state, reward, done, _ = env.step(action)
        ob.append(state, action)
        mask = self.get_mask(env, done)
        self.buffer.append(action, reward, mask, state, done)

        if done:
            self.episode_step = 0
            state = env.reset()
            ob.reset_episode(state)
            self.buffer.reset_episode(state)

        return None

    def select_action(self, ob):
        feature_action = self._preprocess(self.params_model, ob.state, ob.action)
        action = self._select_action(self.params_actor, feature_action)
        return np.array(action[0])

    def explore(self, ob):
        feature_action = self._preprocess(self.params_model, ob.state, ob.action)
        action = self._explore(self.params_actor, feature_action, next(self.rng))[0]
        return np.array(action[0])

    def update(self, writer):
        NotImplementedError

    @abstractmethod
    def update_model(self, writer):
        pass

    @abstractmethod
    def update_sac(self, writer):
        pass

    def save_params(self, save_dir):
        save_params(self.params_model, os.path.join(save_dir, "params_model.npz"))

    def load_params(self, save_dir):
        self.params_model = load_params(os.path.join(save_dir, "params_model.npz"))



from .ddpg import DDPG
from .dqn import DQN
from .fqf import FQF
from .iqn import IQN
from .misc import SlacObservation
from .ppo import PPO
from .qrdqn import QRDQN
from .sac import SAC
from .sac_ae import SAC_AE
from .sac_discor import SAC_DisCor
from .sac_discrete import SAC_Discrete
from .slac import SLAC
from .td3 import TD3
from .tqc import TQC



import math
from functools import partial
from typing import Tuple

import jax
import jax.numpy as jnp
import numpy as np
from jax import nn


@jax.jit
def gaussian_log_prob(
    log_std: jnp.ndarray,
    noise: jnp.ndarray,
) -> jnp.ndarray:
    """
    Calculate log probabilities of gaussian distributions.
    """
    return -0.5 * (jnp.square(noise) + 2 * log_std + jnp.log(2 * math.pi))


@jax.jit
def gaussian_and_tanh_log_prob(
    log_std: jnp.ndarray,
    noise: jnp.ndarray,
    action: jnp.ndarray,
) -> jnp.ndarray:
    """
    Calculate log probabilities of gaussian distributions and tanh transformation.
    """
    return gaussian_log_prob(log_std, noise) - jnp.log(nn.relu(1.0 - jnp.square(action)) + 1e-6)


@jax.jit
def evaluate_gaussian_and_tanh_log_prob(
    mean: jnp.ndarray,
    log_std: jnp.ndarray,
    action: jnp.ndarray,
) -> jnp.ndarray:
    """
    Calculate log probabilities of gaussian distributions and tanh transformation given samples.
    """
    noise = (jnp.arctanh(action) - mean) / (jnp.exp(log_std) + 1e-8)
    return gaussian_and_tanh_log_prob(log_std, noise, action).sum(axis=1, keepdims=True)


@partial(jax.jit, static_argnums=3)
def reparameterize_gaussian(
    mean: jnp.ndarray,
    log_std: jnp.ndarray,
    key: jnp.ndarray,
    return_log_pi: bool = True,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """
    Sample from gaussian distributions.
    """
    std = jnp.exp(log_std)
    noise = jax.random.normal(key, std.shape)
    action = mean + noise * std
    if return_log_pi:
        return action, gaussian_log_prob(log_std, noise).sum(axis=1, keepdims=True)
    else:
        return action


@partial(jax.jit, static_argnums=3)
def reparameterize_gaussian_and_tanh(
    mean: jnp.ndarray,
    log_std: jnp.ndarray,
    key: jnp.ndarray,
    return_log_pi: bool = True,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """
    Sample from gaussian distributions and tanh transforamation.
    """
    std = jnp.exp(log_std)
    noise = jax.random.normal(key, std.shape)
    action = jnp.tanh(mean + noise * std)
    if return_log_pi:
        return action, gaussian_and_tanh_log_prob(log_std, noise, action).sum(axis=1, keepdims=True)
    else:
        return action


@jax.jit
def calculate_kl_divergence(
    p_mean: np.ndarray,
    p_std: np.ndarray,
    q_mean: np.ndarray,
    q_std: np.ndarray,
) -> jnp.ndarray:
    """
    Calculate KL Divergence between gaussian distributions.
    """
    var_ratio = jnp.square(p_std / q_std)
    t1 = jnp.square((p_mean - q_mean) / q_std)
    return 0.5 * (var_ratio + t1 - 1 - jnp.log(var_ratio))



from functools import partial
from typing import List, Tuple

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np

from rljax.algorithm.qrdqn import QRDQN
from rljax.network import DiscreteImplicitQuantileFunction
from rljax.util import fake_state


class IQN(QRDQN):
    name = "IQN"

    def __init__(
        self,
        num_agent_steps,
        state_space,
        action_space,
        seed,
        max_grad_norm=None,
        gamma=0.99,
        nstep=1,
        buffer_size=10 ** 6,
        use_per=False,
        batch_size=32,
        start_steps=50000,
        update_interval=4,
        update_interval_target=8000,
        eps=0.01,
        eps_eval=0.001,
        eps_decay_steps=250000,
        loss_type="huber",
        dueling_net=False,
        double_q=False,
        setup_net=True,
        fn=None,
        lr=5e-5,
        units=(512,),
        num_quantiles=64,
        num_quantiles_eval=32,
        num_cosines=64,
    ):
        if fn is None:

            def fn(s, cum_p):
                return DiscreteImplicitQuantileFunction(
                    num_cosines=num_cosines,
                    action_space=action_space,
                    hidden_units=units,
                    dueling_net=dueling_net,
                )(s, cum_p)

        if not hasattr(self, "fake_args"):
            self.fake_args = (fake_state(state_space), np.empty((1, num_quantiles), dtype=np.float32))
        if not hasattr(self, "use_key_forward"):
            self.use_key_forward = True
        if not hasattr(self, "num_keys_loss"):
            self.num_keys_loss = 3

        super(IQN, self).__init__(
            num_agent_steps=num_agent_steps,
            state_space=state_space,
            action_space=action_space,
            seed=seed,
            max_grad_norm=max_grad_norm,
            gamma=gamma,
            nstep=nstep,
            buffer_size=buffer_size,
            batch_size=batch_size,
            use_per=use_per,
            start_steps=start_steps,
            update_interval=update_interval,
            update_interval_target=update_interval_target,
            eps=eps,
            eps_eval=eps_eval,
            eps_decay_steps=eps_decay_steps,
            loss_type=loss_type,
            dueling_net=dueling_net,
            double_q=double_q,
            setup_net=setup_net,
            fn=fn,
            lr=lr,
            num_quantiles=num_quantiles,
        )
        self.num_quantiles_eval = num_quantiles_eval

    @partial(jax.jit, static_argnums=0)
    def _forward(
        self,
        params: hk.Params,
        state: np.ndarray,
        key: jnp.ndarray,
    ) -> jnp.ndarray:
        cum_p = jax.random.uniform(key, (state.shape[0], self.num_quantiles_eval))
        return jnp.argmax(self.net.apply(params, state, cum_p).mean(axis=1), axis=1)

    @partial(jax.jit, static_argnums=0)
    def _calculate_target(
        self,
        params: hk.Params,
        params_target: hk.Params,
        reward: np.ndarray,
        done: np.ndarray,
        next_state: np.ndarray,
        cum_p_prime: jnp.ndarray,
        key: jnp.ndarray,
    ) -> jnp.ndarray:
        if self.double_q:
            next_action = self._forward(params, next_state, key=key)[..., None]
            next_quantile = self._calculate_value(params_target, next_state, next_action, cum_p=cum_p_prime)
        else:
            next_quantile = jnp.max(self.net.apply(params_target, next_state, cum_p_prime), axis=-1, keepdims=True)
        target = reward[:, None] + (1.0 - done[:, None]) * self.discount * next_quantile
        return jax.lax.stop_gradient(target).reshape(-1, 1, self.num_quantiles)

    @partial(jax.jit, static_argnums=0)
    def _loss(
        self,
        params: hk.Params,
        params_target: hk.Params,
        state: np.ndarray,
        action: np.ndarray,
        reward: np.ndarray,
        done: np.ndarray,
        next_state: np.ndarray,
        weight: np.ndarray,
        key_list: List[np.ndarray],
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        cum_p = jax.random.uniform(key_list[0], (state.shape[0], self.num_quantiles))
        cum_p_prime = jax.random.uniform(key_list[1], (state.shape[0], self.num_quantiles))
        quantile = self._calculate_value(params, state, action, cum_p=cum_p)
        target = self._calculate_target(params, params_target, reward, done, next_state, cum_p_prime, key_list[2])
        return self._calculate_loss_and_abs_td(quantile, target, cum_p, weight)



from functools import partial
from typing import List

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np

from rljax.algorithm.sac import SAC
from rljax.network import ContinuousQuantileFunction, StateDependentGaussianPolicy
from rljax.util import quantile_loss


class TQC(SAC):
    name = "TQC"

    def __init__(
        self,
        num_agent_steps,
        state_space,
        action_space,
        seed,
        max_grad_norm=None,
        gamma=0.99,
        nstep=1,
        num_critics=5,
        buffer_size=10 ** 6,
        use_per=False,
        batch_size=256,
        start_steps=10000,
        update_interval=1,
        tau=5e-3,
        fn_actor=None,
        fn_critic=None,
        lr_actor=3e-4,
        lr_critic=3e-4,
        lr_alpha=3e-4,
        units_actor=(256, 256),
        units_critic=(512, 512, 512),
        log_std_min=-20.0,
        log_std_max=2.0,
        d2rl=False,
        num_quantiles=25,
        num_quantiles_to_drop=0,
    ):
        if d2rl:
            self.name += "-D2RL"

        if fn_critic is None:

            def fn_critic(s, a):
                return ContinuousQuantileFunction(
                    num_critics=num_critics,
                    hidden_units=units_critic,
                    num_quantiles=num_quantiles,
                    d2rl=d2rl,
                )(s, a)

        if fn_actor is None:

            def fn_actor(s):
                return StateDependentGaussianPolicy(
                    action_space=action_space,
                    hidden_units=units_actor,
                    log_std_min=log_std_min,
                    log_std_max=log_std_max,
                    d2rl=d2rl,
                )(s)

        super(TQC, self).__init__(
            num_agent_steps=num_agent_steps,
            state_space=state_space,
            action_space=action_space,
            seed=seed,
            max_grad_norm=max_grad_norm,
            gamma=gamma,
            nstep=nstep,
            num_critics=num_critics,
            buffer_size=buffer_size,
            use_per=use_per,
            batch_size=batch_size,
            start_steps=start_steps,
            update_interval=update_interval,
            tau=tau,
            fn_actor=fn_actor,
            fn_critic=fn_critic,
            lr_actor=lr_actor,
            lr_critic=lr_critic,
            lr_alpha=lr_alpha,
        )
        self.cum_p_prime = jnp.expand_dims((jnp.arange(0, num_quantiles, dtype=jnp.float32) + 0.5) / num_quantiles, 0)
        self.num_quantiles = num_quantiles
        self.num_quantiles_target = num_quantiles * num_critics - num_quantiles_to_drop

    @partial(jax.jit, static_argnums=0)
    def _calculate_value(
        self,
        params_critic: hk.Params,
        state: np.ndarray,
        action: np.ndarray,
    ) -> jnp.ndarray:
        return jnp.concatenate(self._calculate_value_list(params_critic, state, action), axis=1)

    @partial(jax.jit, static_argnums=0)
    def _calculate_target(
        self,
        params_critic_target: hk.Params,
        log_alpha: jnp.ndarray,
        reward: np.ndarray,
        done: np.ndarray,
        next_state: np.ndarray,
        next_action: jnp.ndarray,
        next_log_pi: jnp.ndarray,
    ) -> jnp.ndarray:
        next_quantile = self._calculate_value(params_critic_target, next_state, next_action)
        next_quantile = jnp.sort(next_quantile)[:, : self.num_quantiles_target]
        next_quantile -= jnp.exp(log_alpha) * self._calculate_log_pi(next_action, next_log_pi)
        return jax.lax.stop_gradient(reward + (1.0 - done) * self.discount * next_quantile)

    @partial(jax.jit, static_argnums=0)
    def _calculate_loss_critic_and_abs_td(
        self,
        quantile_list: List[jnp.ndarray],
        target: jnp.ndarray,
        weight: np.ndarray,
    ) -> jnp.ndarray:
        loss_critic = 0.0
        for quantile in quantile_list:
            loss_critic += quantile_loss(target[:, None, :] - quantile[:, :, None], self.cum_p_prime, weight, "huber")
        loss_critic /= self.num_critics * self.num_quantiles
        abs_td = jnp.abs(target[:, None, :] - quantile_list[0][:, :, None]).mean(axis=1).mean(axis=1, keepdims=True)
        return loss_critic, jax.lax.stop_gradient(abs_td)



import jax
import jax.numpy as jnp
import numpy as np


@jax.jit
def add_noise(
    x: jnp.ndarray,
    key: jnp.ndarray,
    std: float,
    out_min: float = -np.inf,
    out_max: float = np.inf,
    noise_min: float = -np.inf,
    noise_max: float = np.inf,
) -> jnp.ndarray:
    """
    Add noise to actions.
    """
    noise = jnp.clip(jax.random.normal(key, x.shape), noise_min, noise_max)
    return jnp.clip(x + noise * std, out_min, out_max)


@jax.jit
def preprocess_state(
    state: np.ndarray,
    key: jnp.ndarray,
) -> jnp.ndarray:
    """
    Preprocess pixel states to fit into [-0.5, 0.5].
    """
    state = state.astype(jnp.float32)
    state = jnp.floor(state / 8)
    state = state / 32
    state = state + jax.random.uniform(key, state.shape) / 32
    state = state - 0.5
    return state


@jax.jit
def get_q_at_action(
    q_s: jnp.ndarray,
    action: jnp.ndarray,
) -> jnp.ndarray:
    """
    Get q values at (s, a).
    """

    def _get(q_s, action):
        return q_s[action]

    return jax.vmap(_get)(q_s, action)


@jax.jit
def get_quantile_at_action(
    quantile_s: jnp.ndarray,
    action: jnp.ndarray,
) -> jnp.ndarray:
    """
    Get quantile values at (s, a).
    """

    def _get(quantile_s, action):
        return quantile_s[:, action]

    return jax.vmap(_get)(quantile_s, action)



import haiku as hk
import jax.numpy as jnp
from jax import nn

from rljax.network.base import MLP
from rljax.network.conv import DQNBody


class DeterministicPolicy(hk.Module):
    """
    Policy for DDPG and TD3.
    """

    def __init__(
        self,
        action_space,
        hidden_units=(256, 256),
        d2rl=False,
    ):
        super(DeterministicPolicy, self).__init__()
        self.action_space = action_space
        self.hidden_units = hidden_units
        self.d2rl = d2rl

    def __call__(self, x):
        return MLP(
            self.action_space.shape[0],
            self.hidden_units,
            hidden_activation=nn.relu,
            output_activation=jnp.tanh,
            d2rl=self.d2rl,
        )(x)


class StateDependentGaussianPolicy(hk.Module):
    """
    Policy for SAC.
    """

    def __init__(
        self,
        action_space,
        hidden_units=(256, 256),
        log_std_min=-20.0,
        log_std_max=2.0,
        clip_log_std=True,
        d2rl=False,
    ):
        super(StateDependentGaussianPolicy, self).__init__()
        self.action_space = action_space
        self.hidden_units = hidden_units
        self.log_std_min = log_std_min
        self.log_std_max = log_std_max
        self.clip_log_std = clip_log_std
        self.d2rl = d2rl

    def __call__(self, x):
        x = MLP(
            2 * self.action_space.shape[0],
            self.hidden_units,
            hidden_activation=nn.relu,
            d2rl=self.d2rl,
        )(x)
        mean, log_std = jnp.split(x, 2, axis=1)
        if self.clip_log_std:
            log_std = jnp.clip(log_std, self.log_std_min, self.log_std_max)
        else:
            log_std = self.log_std_min + 0.5 * (self.log_std_max - self.log_std_min) * (jnp.tanh(log_std) + 1.0)
        return mean, log_std


class StateIndependentGaussianPolicy(hk.Module):
    """
    Policy for PPO.
    """

    def __init__(
        self,
        action_space,
        hidden_units=(64, 64),
    ):
        super(StateIndependentGaussianPolicy, self).__init__()
        self.action_space = action_space
        self.hidden_units = hidden_units

    def __call__(self, x):
        mean = MLP(
            self.action_space.shape[0],
            self.hidden_units,
            hidden_activation=jnp.tanh,
            output_scale=0.01,
        )(x)
        log_std = hk.get_parameter("log_std", (1, self.action_space.shape[0]), init=jnp.zeros)
        return mean, log_std


class CategoricalPolicy(hk.Module):
    """
    Policy for SAC-Discrete.
    """

    def __init__(
        self,
        action_space,
        hidden_units=(512,),
    ):
        super(CategoricalPolicy, self).__init__()
        self.action_space = action_space
        self.hidden_units = hidden_units

    def __call__(self, x):
        if len(x.shape) == 4:
            x = DQNBody()(x)
        x = MLP(
            self.action_space.n,
            self.hidden_units,
            hidden_activation=nn.relu,
            output_scale=0.01,
        )(x)
        pi_s = nn.softmax(x, axis=1)
        log_pi_s = jnp.log(pi_s + (pi_s == 0.0) * 1e-8)
        return pi_s, log_pi_s



# NOTE: this code was mainly taken from:
# https://github.com/denisyarats/dmc2gym/blob/master/dmc2gym/wrappers.py
import gym
import numpy as np
from dm_control import suite
from gym import core, spaces

from dm_env import specs
from rljax.env.atari import FrameStack

gym.logger.set_level(40)


def make_dmc_env(domain_name, task_name, action_repeat, n_frames=3, image_size=84):
    env = make(
        domain_name=domain_name,
        task_name=task_name,
        visualize_reward=False,
        from_pixels=True,
        height=image_size,
        width=image_size,
        frame_skip=action_repeat,
    )
    if n_frames != 1:
        env = FrameStack(env, n_frames=n_frames)
    if not hasattr(env, "_max_episode_steps"):
        setattr(env, "_max_episode_steps", env.env._max_episode_steps)
    return env


def _spec_to_box(spec):
    def extract_min_max(s):
        assert s.dtype == np.float64 or s.dtype == np.float32
        dim = np.int(np.prod(s.shape))
        if type(s) == specs.Array:
            bound = np.inf * np.ones(dim, dtype=np.float32)
            return -bound, bound
        elif type(s) == specs.BoundedArray:
            zeros = np.zeros(dim, dtype=np.float32)
            return s.minimum + zeros, s.maximum + zeros

    mins, maxs = [], []
    for s in spec:
        mn, mx = extract_min_max(s)
        mins.append(mn)
        maxs.append(mx)
    low = np.concatenate(mins, axis=0)
    high = np.concatenate(maxs, axis=0)
    assert low.shape == high.shape
    return spaces.Box(low, high, dtype=np.float32)


def _flatten_obs(obs):
    obs_pieces = []
    for v in obs.values():
        flat = np.array([v]) if np.isscalar(v) else v.ravel()
        obs_pieces.append(flat)
    return np.concatenate(obs_pieces, axis=0)


def make(
    domain_name,
    task_name,
    seed=1,
    visualize_reward=True,
    from_pixels=False,
    height=84,
    width=84,
    camera_id=0,
    frame_skip=1,
    episode_length=1000,
    environment_kwargs=None,
    time_limit=None,
):
    env_id = "dmc_%s_%s_%s_%s_%s-v1" % (domain_name, task_name, seed, height, width)

    if from_pixels:
        assert not visualize_reward, "cannot use visualize reward when learning from pixels"

    # shorten episode length
    max_episode_steps = (episode_length + frame_skip - 1) // frame_skip

    if env_id not in gym.envs.registry.env_specs:
        task_kwargs = {}
        if seed is not None:
            task_kwargs["random"] = seed
        if time_limit is not None:
            task_kwargs["time_limit"] = time_limit
        gym.envs.registration.register(
            id=env_id,
            entry_point=DMCWrapper,
            kwargs=dict(
                domain_name=domain_name,
                task_name=task_name,
                task_kwargs=task_kwargs,
                environment_kwargs=environment_kwargs,
                visualize_reward=visualize_reward,
                from_pixels=from_pixels,
                height=height,
                width=width,
                camera_id=camera_id,
                frame_skip=frame_skip,
            ),
            max_episode_steps=max_episode_steps,
        )
    return gym.make(env_id)


class DMCWrapper(core.Env):
    def __init__(
        self,
        domain_name,
        task_name,
        task_kwargs=None,
        visualize_reward={},
        from_pixels=False,
        height=84,
        width=84,
        camera_id=0,
        frame_skip=1,
        environment_kwargs=None,
    ):
        assert "random" in task_kwargs, "please specify a seed, for deterministic behaviour"
        self._from_pixels = from_pixels
        self._height = height
        self._width = width
        self._camera_id = camera_id
        self._frame_skip = frame_skip

        # Create task.
        self._env = suite.load(
            domain_name=domain_name,
            task_name=task_name,
            task_kwargs=task_kwargs,
            visualize_reward=visualize_reward,
            environment_kwargs=environment_kwargs,
        )

        # True action space.
        self._true_action_space = _spec_to_box([self._env.action_spec()])
        self._low = self._true_action_space.low
        self._delta = self._true_action_space.high - self._true_action_space.low
        # Normalized action space.
        self._norm_action_space = spaces.Box(low=-1.0, high=1.0, shape=self._true_action_space.shape, dtype=np.float64)

        # Create observation space.
        if from_pixels:
            self._observation_space = spaces.Box(low=0, high=255, shape=[height, width, 3], dtype=np.uint8)
        else:
            self._observation_space = _spec_to_box(self._env.observation_spec().values())

        self._state_space = _spec_to_box(self._env.observation_spec().values())
        self.current_state = None

        # Set seed.
        self.seed(seed=task_kwargs.get("random", 1))

    def __getattr__(self, name):
        return getattr(self._env, name)

    def _get_obs(self, time_step):
        if self._from_pixels:
            obs = self.render(height=self._height, width=self._width, camera_id=self._camera_id)
        else:
            obs = _flatten_obs(time_step.observation)
        return obs

    def _convert_action(self, action):
        return (action + 1.0) * self._delta / 2.0 + self._low

    @property
    def observation_space(self):
        return self._observation_space

    @property
    def state_space(self):
        return self._state_space

    @property
    def action_space(self):
        return self._norm_action_space

    def seed(self, seed):
        self._true_action_space.seed(seed)
        self._norm_action_space.seed(seed)
        self._observation_space.seed(seed)

    def step(self, action):
        action = np.clip(
            action,
            a_min=self._norm_action_space.low,
            a_max=self._norm_action_space.high,
        )
        assert self._norm_action_space.contains(action)
        action = self._convert_action(action)
        assert self._true_action_space.contains(action)
        reward = 0.0

        for _ in range(self._frame_skip):
            time_step = self._env.step(action)
            reward += time_step.reward or 0.0
            done = time_step.last()
            if done:
                break
        obs = self._get_obs(time_step)
        self.current_state = _flatten_obs(time_step.observation)
        return obs, reward, done, {}

    def reset(self):
        time_step = self._env.reset()
        self.current_state = _flatten_obs(time_step.observation)
        return self._get_obs(time_step)

    def render(self, mode="rgb_array", height=None, width=None, camera_id=0):
        assert mode == "rgb_array", "only support rgb_array mode, given %s" % mode
        height = height or self._height
        width = width or self._width
        camera_id = camera_id or self._camera_id
        return self._env.physics.render(height=height, width=width, camera_id=camera_id)



import os
from functools import partial
from typing import Tuple

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
from jax.experimental import optix

from rljax.algorithm.qrdqn import QRDQN
from rljax.network import CumProbNetwork, DiscreteImplicitQuantileFunction, make_quantile_nerwork
from rljax.util import get_quantile_at_action, load_params, optimize, save_params


class FQF(QRDQN):
    name = "FQF"

    def __init__(
        self,
        num_agent_steps,
        state_space,
        action_space,
        seed,
        max_grad_norm=None,
        gamma=0.99,
        nstep=1,
        buffer_size=10 ** 6,
        use_per=False,
        batch_size=32,
        start_steps=50000,
        update_interval=4,
        update_interval_target=8000,
        eps=0.01,
        eps_eval=0.001,
        eps_decay_steps=250000,
        loss_type="huber",
        dueling_net=False,
        double_q=False,
        setup_net=True,
        fn=None,
        lr=5e-5,
        lr_cum_p=2.5e-9,
        units=(512,),
        num_quantiles=32,
        num_cosines=64,
    ):
        super(FQF, self).__init__(
            num_agent_steps=num_agent_steps,
            state_space=state_space,
            action_space=action_space,
            seed=seed,
            max_grad_norm=max_grad_norm,
            gamma=gamma,
            nstep=nstep,
            buffer_size=buffer_size,
            batch_size=batch_size,
            use_per=use_per,
            start_steps=start_steps,
            update_interval=update_interval,
            update_interval_target=update_interval_target,
            eps=eps,
            eps_eval=eps_eval,
            eps_decay_steps=eps_decay_steps,
            loss_type=loss_type,
            dueling_net=dueling_net,
            double_q=double_q,
            setup_net=False,
            num_quantiles=num_quantiles,
        )
        if setup_net:
            if fn is None:

                def fn(s, cum_p):
                    return DiscreteImplicitQuantileFunction(
                        num_cosines=num_cosines,
                        action_space=action_space,
                        hidden_units=units,
                        dueling_net=dueling_net,
                    )(s, cum_p)

            self.net, self.params, fake_feature = make_quantile_nerwork(self.rng, state_space, action_space, fn, num_quantiles)
            self.params_target = self.params
            opt_init, self.opt = optix.adam(lr, eps=0.01 / batch_size)
            self.opt_state = opt_init(self.params)

        # Fraction proposal network.
        self.cum_p_net = hk.without_apply_rng(hk.transform(lambda s: CumProbNetwork(num_quantiles=num_quantiles)(s)))
        self.params_cum_p = self.cum_p_net.init(next(self.rng), fake_feature)
        opt_init, self.opt_cum_p = optix.rmsprop(lr_cum_p, decay=0.95, eps=1e-5, centered=True)
        self.opt_state_cum_p = opt_init(self.params_cum_p)

    def forward(self, state):
        return self._forward(self.params_cum_p, self.params, state)

    @partial(jax.jit, static_argnums=0)
    def _forward(
        self,
        params_cum_p: hk.Params,
        params: hk.Params,
        state: np.ndarray,
    ) -> jnp.ndarray:
        feature = self.net["feature"].apply(params["feature"], state)
        return self._forward_from_feature(params_cum_p, params, feature)

    @partial(jax.jit, static_argnums=0)
    def _forward_from_feature(
        self,
        params_cum_p: hk.Params,
        params: hk.Params,
        feature: np.ndarray,
    ) -> jnp.ndarray:
        cum_p, cum_p_prime = self.cum_p_net.apply(params_cum_p, feature)
        quantile_s = self.net["quantile"].apply(params["quantile"], feature, cum_p_prime)
        q_s = ((cum_p[:, 1:, None] - cum_p[:, :-1, None]) * quantile_s).sum(axis=1)
        return jnp.argmax(q_s, axis=1)

    def update(self, writer=None):
        self.learning_step += 1
        weight, batch = self.buffer.sample(self.batch_size)
        state, action, reward, done, next_state = batch

        # Update fraction proposal network.
        self.opt_state_cum_p, self.params_cum_p, loss_cum_p, _ = optimize(
            self._loss_cum_p,
            self.opt_cum_p,
            self.opt_state_cum_p,
            self.params_cum_p,
            self.max_grad_norm,
            params=self.params,
            state=state,
            action=action,
        )

        # Update quantile network.
        self.opt_state, self.params, loss, abs_td = optimize(
            self._loss,
            self.opt,
            self.opt_state,
            self.params,
            self.max_grad_norm,
            params_target=self.params_target,
            params_cum_p=self.params_cum_p,
            state=state,
            action=action,
            reward=reward,
            done=done,
            next_state=next_state,
            weight=weight,
            **self.kwargs_update,
        )

        # Update priority.
        if self.use_per:
            self.buffer.update_priority(abs_td)

        # Update target network.
        if self.agent_step % self.update_interval_target == 0:
            self.params_target = self._update_target(self.params_target, self.params)

        if writer and self.learning_step % 1000 == 0:
            writer.add_scalar("loss/q", loss, self.learning_step)
            writer.add_scalar("loss/cum_p", loss_cum_p, self.learning_step)

    @partial(jax.jit, static_argnums=0)
    def _calculate_value(
        self,
        params: hk.Params,
        feature: np.ndarray,
        action: np.ndarray,
        cum_p: jnp.ndarray,
    ) -> jnp.ndarray:
        return get_quantile_at_action(self.net["quantile"].apply(params["quantile"], feature, cum_p), action)

    @partial(jax.jit, static_argnums=0)
    def _calculate_target(
        self,
        params_cum_p: hk.Params,
        params: hk.Params,
        params_target: hk.Params,
        reward: np.ndarray,
        done: np.ndarray,
        next_feature: np.ndarray,
        cum_p_prime: jnp.ndarray,
    ) -> jnp.ndarray:
        if self.double_q:
            next_action = self._forward_from_feature(params_cum_p, params, next_feature)[:, None]
            next_quantile = self._calculate_value(params_target, next_feature, next_action, cum_p_prime)
        else:
            next_quantile_s = self.net["quantile"].apply(params_target["quantile"], next_feature, cum_p_prime)
            next_quantile = jnp.max(next_quantile_s, axis=-1, keepdims=True)
        target = reward[:, None] + (1.0 - done[:, None]) * self.discount * next_quantile
        return jax.lax.stop_gradient(target).reshape(-1, 1, self.num_quantiles)

    @partial(jax.jit, static_argnums=0)
    def _loss(
        self,
        params: hk.Params,
        params_target: hk.Params,
        params_cum_p: hk.Params,
        state: np.ndarray,
        action: np.ndarray,
        reward: np.ndarray,
        done: np.ndarray,
        next_state: np.ndarray,
        weight: np.ndarray,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        feature = self.net["feature"].apply(params["feature"], state)
        next_feature = self.net["feature"].apply(params_target["feature"], next_state)
        cum_p_prime = jax.lax.stop_gradient(self.cum_p_net.apply(params_cum_p, feature)[1])
        quantile = self._calculate_value(params, feature, action, cum_p_prime)
        target = self._calculate_target(params_cum_p, params, params_target, reward, done, next_feature, cum_p_prime)
        return self._calculate_loss_and_abs_td(quantile, target, cum_p_prime, weight)

    @partial(jax.jit, static_argnums=0)
    def _loss_cum_p(self, params_cum_p, params, state, action):
        feature = jax.lax.stop_gradient(self.net["feature"].apply(params["feature"], state))
        cum_p, cum_p_prime = self.cum_p_net.apply(params_cum_p, feature)
        quantile = get_quantile_at_action(self.net["quantile"].apply(params["quantile"], feature, cum_p[:, 1:-1]), action)
        quantile_prime = get_quantile_at_action(self.net["quantile"].apply(params["quantile"], feature, cum_p_prime), action)
        # NOTE: Proposition 1 in the paper requires F^{-1} is non-decreasing. I relax this requirements and
        # calculate gradients of taus even when F^{-1} is not non-decreasing.
        val1 = quantile - quantile_prime[:, :-1]
        sign1 = quantile > jnp.concatenate([quantile_prime[:, :1], quantile[:, :-1]], axis=1)
        val2 = quantile - quantile_prime[:, 1:]
        sign2 = quantile < jnp.concatenate([quantile[:, 1:], quantile_prime[:, -1:]], axis=1)
        grad = jnp.where(sign1, val1, -val1) + jnp.where(sign2, val2, -val2)
        grad = jax.lax.stop_gradient(grad.reshape(-1, self.num_quantiles - 1))
        return (cum_p[:, 1:-1] * grad).sum(axis=1).mean(), None

    def save_params(self, save_dir):
        super().save_params(save_dir)
        save_params(self.params_cum_p, os.path.join(save_dir, "params_cum_p.npz"))

    def load_params(self, save_dir):
        super().load_params(save_dir)
        self.params_cum_p = load_params(os.path.join(save_dir, "params_cum_p.npz"))



from functools import partial
from typing import Tuple

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
from jax.experimental import optix

from rljax.algorithm.base_class import OnPolicyActorCritic
from rljax.network import ContinuousVFunction, StateIndependentGaussianPolicy
from rljax.util import evaluate_gaussian_and_tanh_log_prob, optimize, reparameterize_gaussian_and_tanh


class PPO(OnPolicyActorCritic):
    name = "PPO"

    def __init__(
        self,
        num_agent_steps,
        state_space,
        action_space,
        seed,
        max_grad_norm=10.0,
        gamma=0.995,
        buffer_size=2048,
        batch_size=64,
        fn_actor=None,
        fn_critic=None,
        lr_actor=3e-4,
        lr_critic=3e-4,
        units_actor=(64, 64),
        units_critic=(64, 64),
        epoch_ppo=10,
        clip_eps=0.2,
        lambd=0.97,
    ):
        assert buffer_size % batch_size == 0
        super(PPO, self).__init__(
            num_agent_steps=num_agent_steps,
            state_space=state_space,
            action_space=action_space,
            seed=seed,
            max_grad_norm=max_grad_norm,
            gamma=gamma,
            buffer_size=buffer_size,
            batch_size=batch_size,
        )

        if fn_critic is None:

            def fn_critic(s):
                return ContinuousVFunction(
                    hidden_units=(units_critic),
                )(s)

        if fn_actor is None:

            def fn_actor(s):
                return StateIndependentGaussianPolicy(
                    action_space=action_space,
                    hidden_units=units_actor,
                )(s)

        # Critic.
        self.critic = hk.without_apply_rng(hk.transform(fn_critic))
        self.params_critic = self.params_critic_target = self.critic.init(next(self.rng), *self.fake_args_critic)
        opt_init, self.opt_critic = optix.adam(lr_critic)
        self.opt_state_critic = opt_init(self.params_critic)

        # Actor.
        self.actor = hk.without_apply_rng(hk.transform(fn_actor))
        self.params_actor = self.params_actor_target = self.actor.init(next(self.rng), *self.fake_args_actor)
        opt_init, self.opt_actor = optix.adam(lr_actor)
        self.opt_state_actor = opt_init(self.params_actor)

        # Other parameters.
        self.epoch_ppo = epoch_ppo
        self.clip_eps = clip_eps
        self.lambd = lambd
        self.max_grad_norm = max_grad_norm
        self.idxes = np.arange(buffer_size)

    @partial(jax.jit, static_argnums=0)
    def _select_action(
        self,
        params_actor: hk.Params,
        state: np.ndarray,
    ) -> jnp.ndarray:
        mean, _ = self.actor.apply(params_actor, state)
        return jnp.tanh(mean)

    @partial(jax.jit, static_argnums=0)
    def _explore(
        self,
        params_actor: hk.Params,
        state: np.ndarray,
        key: jnp.ndarray,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        mean, log_std = self.actor.apply(params_actor, state)
        return reparameterize_gaussian_and_tanh(mean, log_std, key, True)

    def update(self, writer=None):
        state, action, reward, done, log_pi_old, next_state = self.buffer.get()

        # Calculate gamma-returns and GAEs.
        gae, target = self.calculate_gae(
            params_critic=self.params_critic,
            state=state,
            reward=reward,
            done=done,
            next_state=next_state,
        )

        for _ in range(self.epoch_ppo):
            np.random.shuffle(self.idxes)
            for start in range(0, self.buffer_size, self.batch_size):
                self.learning_step += 1
                idx = self.idxes[start : start + self.batch_size]

                # Update critic.
                self.opt_state_critic, self.params_critic, loss_critic, _ = optimize(
                    self._loss_critic,
                    self.opt_critic,
                    self.opt_state_critic,
                    self.params_critic,
                    self.max_grad_norm,
                    state=state[idx],
                    target=target[idx],
                )

                # Update actor.
                self.opt_state_actor, self.params_actor, loss_actor, _ = optimize(
                    self._loss_actor,
                    self.opt_actor,
                    self.opt_state_actor,
                    self.params_actor,
                    self.max_grad_norm,
                    state=state[idx],
                    action=action[idx],
                    log_pi_old=log_pi_old[idx],
                    gae=gae[idx],
                )

        if writer:
            writer.add_scalar("loss/critic", loss_critic, self.learning_step)
            writer.add_scalar("loss/actor", loss_actor, self.learning_step)

    @partial(jax.jit, static_argnums=0)
    def _loss_critic(
        self,
        params_critic: hk.Params,
        state: np.ndarray,
        target: np.ndarray,
    ) -> jnp.ndarray:
        return jnp.square(target - self.critic.apply(params_critic, state)).mean(), None

    @partial(jax.jit, static_argnums=0)
    def _loss_actor(
        self,
        params_actor: hk.Params,
        state: np.ndarray,
        action: np.ndarray,
        log_pi_old: np.ndarray,
        gae: jnp.ndarray,
    ) -> jnp.ndarray:
        # Calculate log(\pi) at current policy.
        mean, log_std = self.actor.apply(params_actor, state)
        log_pi = evaluate_gaussian_and_tanh_log_prob(mean, log_std, action)
        # Calculate importance ratio.
        ratio = jnp.exp(log_pi - log_pi_old)
        loss_actor1 = -ratio * gae
        loss_actor2 = -jnp.clip(ratio, 1.0 - self.clip_eps, 1.0 + self.clip_eps) * gae
        loss_actor = jnp.maximum(loss_actor1, loss_actor2).mean()
        return loss_actor, None

    @partial(jax.jit, static_argnums=0)
    def calculate_gae(
        self,
        params_critic: hk.Params,
        state: np.ndarray,
        reward: np.ndarray,
        done: np.ndarray,
        next_state: np.ndarray,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        # Current and next value estimates.
        value = jax.lax.stop_gradient(self.critic.apply(params_critic, state))
        next_value = jax.lax.stop_gradient(self.critic.apply(params_critic, next_state))
        # Calculate TD errors.
        delta = reward + self.gamma * next_value * (1.0 - done) - value
        # Calculate GAE recursively from behind.
        gae = [delta[-1]]
        for t in jnp.arange(self.buffer_size - 2, -1, -1):
            gae.insert(0, delta[t] + self.gamma * self.lambd * (1 - done[t]) * gae[0])
        gae = jnp.array(gae)
        return (gae - gae.mean()) / (gae.std() + 1e-8), gae + value



from functools import partial
from typing import Tuple

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np

from rljax.algorithm.dqn import DQN
from rljax.network import DiscreteQuantileFunction
from rljax.util import get_quantile_at_action, quantile_loss


class QRDQN(DQN):
    name = "QR-DQN"

    def __init__(
        self,
        num_agent_steps,
        state_space,
        action_space,
        seed,
        max_grad_norm=None,
        gamma=0.99,
        nstep=1,
        buffer_size=10 ** 6,
        use_per=False,
        batch_size=32,
        start_steps=50000,
        update_interval=4,
        update_interval_target=8000,
        eps=0.01,
        eps_eval=0.001,
        eps_decay_steps=250000,
        loss_type="huber",
        dueling_net=False,
        double_q=False,
        setup_net=True,
        fn=None,
        lr=5e-5,
        units=(512,),
        num_quantiles=200,
    ):
        if fn is None:

            def fn(s):
                return DiscreteQuantileFunction(
                    action_space=action_space,
                    num_quantiles=num_quantiles,
                    hidden_units=units,
                    dueling_net=dueling_net,
                )(s)

        super(QRDQN, self).__init__(
            num_agent_steps=num_agent_steps,
            state_space=state_space,
            action_space=action_space,
            seed=seed,
            max_grad_norm=max_grad_norm,
            gamma=gamma,
            nstep=nstep,
            buffer_size=buffer_size,
            batch_size=batch_size,
            use_per=use_per,
            start_steps=start_steps,
            update_interval=update_interval,
            update_interval_target=update_interval_target,
            eps=eps,
            eps_eval=eps_eval,
            eps_decay_steps=eps_decay_steps,
            loss_type=loss_type,
            dueling_net=dueling_net,
            double_q=double_q,
            setup_net=setup_net,
            fn=fn,
            lr=lr,
        )
        if self.name == "QR-DQN":
            self.cum_p_prime = jnp.expand_dims((jnp.arange(0, num_quantiles, dtype=jnp.float32) + 0.5) / num_quantiles, 0)
        self.num_quantiles = num_quantiles

    @partial(jax.jit, static_argnums=0)
    def _forward(
        self,
        params: hk.Params,
        state: np.ndarray,
    ) -> jnp.ndarray:
        return jnp.argmax(self.net.apply(params, state).mean(axis=1), axis=1)

    @partial(jax.jit, static_argnums=0)
    def _calculate_value(
        self,
        params: hk.Params,
        state: np.ndarray,
        action: np.ndarray,
        *args,
        **kwargs,
    ) -> jnp.ndarray:
        return get_quantile_at_action(self.net.apply(params, state, *args, **kwargs), action)

    @partial(jax.jit, static_argnums=0)
    def _calculate_target(
        self,
        params: hk.Params,
        params_target: hk.Params,
        reward: np.ndarray,
        done: np.ndarray,
        next_state: np.ndarray,
        *args,
        **kwargs,
    ) -> jnp.ndarray:
        if self.double_q:
            next_action = self._forward(params, next_state, *args, **kwargs)[..., None]
            next_quantile = self._calculate_value(params_target, next_state, next_action)
        else:
            next_quantile = jnp.max(self.net.apply(params_target, next_state), axis=-1, keepdims=True)
        target = reward[:, None] + (1.0 - done[:, None]) * self.discount * next_quantile
        return jax.lax.stop_gradient(target).reshape(-1, 1, self.num_quantiles)

    @partial(jax.jit, static_argnums=0)
    def _calculate_loss_and_abs_td(
        self,
        quantile: jnp.ndarray,
        target: jnp.ndarray,
        cum_p: jnp.ndarray,
        weight: np.ndarray,
    ) -> jnp.ndarray:
        td = target - quantile
        loss = quantile_loss(td, cum_p, weight, self.loss_type)
        return loss, jax.lax.stop_gradient(jnp.abs(td).sum(axis=1).mean(axis=1, keepdims=True))

    @partial(jax.jit, static_argnums=0)
    def _loss(
        self,
        params: hk.Params,
        params_target: hk.Params,
        state: np.ndarray,
        action: np.ndarray,
        reward: np.ndarray,
        done: np.ndarray,
        next_state: np.ndarray,
        weight: np.ndarray,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        quantile = self._calculate_value(params, state, action)
        target = self._calculate_target(params, params_target, reward, done, next_state)
        return self._calculate_loss_and_abs_td(quantile, target, self.cum_p_prime, weight)



from functools import partial

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
from jax import nn

from rljax.network.base import MLP
from rljax.network.conv import DQNBody, SLACDecoder, SLACEncoder


class CumProbNetwork(hk.Module):
    """
    Fraction Proposal Network for FQF.
    """

    def __init__(self, num_quantiles=64):
        super(CumProbNetwork, self).__init__()
        self.num_quantiles = num_quantiles

    def __call__(self, x):
        w_init = hk.initializers.Orthogonal(scale=1.0 / np.sqrt(3.0))
        p = nn.softmax(hk.Linear(self.num_quantiles, w_init=w_init)(x))
        cum_p = jnp.concatenate([jnp.zeros((p.shape[0], 1)), jnp.cumsum(p, axis=1)], axis=1)
        cum_p_prime = (cum_p[:, 1:] + cum_p[:, :-1]) / 2.0
        return cum_p, cum_p_prime


def make_quantile_nerwork(
    rng,
    state_space,
    action_space,
    fn,
    num_quantiles,
):
    """
    Make Quantile Nerwork for FQF.
    """
    fake_state = state_space.sample()[None, ...]
    if len(state_space.shape) == 1:
        fake_state = fake_state.astype(np.float32)
    network_dict = {}
    params_dict = {}

    if len(state_space.shape) == 3:
        network_dict["feature"] = hk.without_apply_rng(hk.transform(lambda s: DQNBody()(s)))
        fake_feature = np.zeros((1, 7 * 7 * 64), dtype=np.float32)
    else:
        network_dict["feature"] = hk.without_apply_rng(hk.transform(lambda s: s))
        fake_feature = fake_state
    params_dict["feature"] = network_dict["feature"].init(next(rng), fake_state)

    fake_cum_p = np.empty((1, num_quantiles), dtype=np.float32)
    network_dict["quantile"] = hk.without_apply_rng(hk.transform(fn))
    params_dict["quantile"] = network_dict["quantile"].init(next(rng), fake_feature, fake_cum_p)

    network_dict = hk.data_structures.to_immutable_dict(network_dict)
    params_dict = hk.data_structures.to_immutable_dict(params_dict)
    return network_dict, params_dict, fake_feature


class SACLinear(hk.Module):
    """
    Linear layer for SAC+AE.
    """

    def __init__(self, feature_dim):
        super().__init__()
        self.feature_dim = feature_dim

    def __call__(self, x):
        w_init = hk.initializers.Orthogonal(scale=1.0)
        x = hk.Linear(self.feature_dim, w_init=w_init)(x)
        x = hk.LayerNorm(axis=1, create_scale=True, create_offset=True)(x)
        x = jnp.tanh(x)
        return x


class ConstantGaussian(hk.Module):
    """
    Constant diagonal gaussian distribution for SLAC.
    """

    def __init__(self, output_dim, std):
        super().__init__()
        self.output_dim = output_dim
        self.std = std

    def __call__(self, x):
        mean = jnp.zeros((x.shape[0], self.output_dim))
        std = jnp.ones((x.shape[0], self.output_dim)) * self.std
        return jax.lax.stop_gradient(mean), jax.lax.stop_gradient(std)


class Gaussian(hk.Module):
    """
    Diagonal gaussian distribution with state dependent variances for SLAC.
    """

    def __init__(self, output_dim, hidden_units=(256, 256), negative_slope=0.2):
        super().__init__()
        self.output_dim = output_dim
        self.hidden_units = hidden_units
        self.negative_slope = negative_slope

    def __call__(self, x):
        x = MLP(
            output_dim=2 * self.output_dim,
            hidden_units=self.hidden_units,
            hidden_activation=partial(nn.leaky_relu, negative_slope=self.negative_slope),
        )(x)
        mean, log_std = jnp.split(x, 2, axis=1)
        std = nn.softplus(log_std) + 1e-5
        return mean, std


def make_stochastic_latent_variable_model(
    rng,
    state_space,
    action_space,
    num_sequences,
    units_model,
    z1_dim,
    z2_dim,
    feature_dim,
):
    """
    Make Stochastic Latent Variable Model for SLAC.
    """

    # Fake input for JIT compilations.
    fake_state_ = jnp.empty((1, num_sequences, *state_space.shape), dtype=jnp.uint8)
    fake_action_ = jnp.empty((1, num_sequences, *action_space.shape))
    fake_action = jnp.empty((1, *action_space.shape))
    fake_feature = jnp.empty((1, feature_dim))
    fake_z_ = jnp.empty((1, num_sequences, z1_dim + z2_dim))
    fake_z1_ = jnp.empty((1, num_sequences, z1_dim))
    fake_z2_ = jnp.empty((1, num_sequences, z2_dim))
    fake_z1 = jnp.empty((1, z1_dim))
    fake_z2 = jnp.empty((1, z2_dim))

    def fn_z1_prior(z2, a):
        return Gaussian(output_dim=z1_dim, hidden_units=units_model)(jnp.concatenate([z2, a], axis=1))

    def fn_z1_post(f, z2, a):
        return Gaussian(output_dim=z1_dim, hidden_units=units_model)(jnp.concatenate([f, z2, a], axis=1))

    def fn_z2(z1, z2, a):
        return Gaussian(output_dim=z2_dim, hidden_units=units_model)(jnp.concatenate([z1, z2, a], axis=1))

    def fn_reward(z_, a_, n_z_):
        x = jnp.concatenate([z_, a_, n_z_], axis=-1)
        B, S, X = x.shape
        mean, std = Gaussian(output_dim=1, hidden_units=units_model)(x.reshape([B * S, X]))
        return mean.reshape([B, S, 1]), std.reshape([B, S, 1])

    def fn_encoder(x):
        return SLACEncoder(output_dim=feature_dim)(x)

    def fn_decoder(z1_, z2_):
        return SLACDecoder(state_space=state_space, std=np.sqrt(0.1))(jnp.concatenate([z1_, z2_], axis=-1))

    network_dict = {}
    params_dict = {}

    network_dict["z1_prior_init"] = hk.without_apply_rng(hk.transform(lambda x: ConstantGaussian(z1_dim, 1.0)(x)))
    params_dict["z1_prior_init"] = network_dict["z1_prior_init"].init(next(rng), fake_action)

    network_dict["z1_prior"] = hk.without_apply_rng(hk.transform(fn_z1_prior))
    params_dict["z1_prior"] = network_dict["z1_prior"].init(next(rng), fake_z2, fake_action)

    network_dict["z1_post_init"] = hk.without_apply_rng(hk.transform(lambda x: Gaussian(z1_dim, units_model)(x)))
    params_dict["z1_post_init"] = network_dict["z1_post_init"].init(next(rng), fake_feature)

    network_dict["z1_post"] = hk.without_apply_rng(hk.transform(fn_z1_post))
    params_dict["z1_post"] = network_dict["z1_post"].init(next(rng), fake_feature, fake_z2, fake_action)

    network_dict["z2_init"] = hk.without_apply_rng(hk.transform(lambda x: Gaussian(z2_dim, units_model)(x)))
    params_dict["z2_init"] = network_dict["z2_init"].init(next(rng), fake_z1)

    network_dict["z2"] = hk.without_apply_rng(hk.transform(fn_z2))
    params_dict["z2"] = network_dict["z2"].init(next(rng), fake_z1, fake_z2, fake_action)

    network_dict["reward"] = hk.without_apply_rng(hk.transform(fn_reward))
    params_dict["reward"] = network_dict["reward"].init(next(rng), fake_z_, fake_action_, fake_z_)

    network_dict["encoder"] = hk.without_apply_rng(hk.transform(fn_encoder))
    params_dict["encoder"] = network_dict["encoder"].init(next(rng), fake_state_)

    network_dict["decoder"] = hk.without_apply_rng(hk.transform(fn_decoder))
    params_dict["decoder"] = network_dict["decoder"].init(next(rng), fake_z1_, fake_z2_)

    network_dict = hk.data_structures.to_immutable_dict(network_dict)
    params_dict = hk.data_structures.to_immutable_dict(params_dict)
    return network_dict, params_dict



import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
from jax import nn

from rljax.network.initializer import DeltaOrthogonal


class DQNBody(hk.Module):
    """
    CNN for the atari environment.
    """

    def __init__(self):
        super(DQNBody, self).__init__()

    def __call__(self, x):
        # He's initializer.
        w_init = hk.initializers.Orthogonal(scale=np.sqrt(2))
        # Floatify the image.
        x = x.astype(jnp.float32) / 255.0
        # Apply CNN.
        x = hk.Conv2D(32, kernel_shape=8, stride=4, padding="VALID", w_init=w_init)(x)
        x = nn.relu(x)
        x = hk.Conv2D(64, kernel_shape=4, stride=2, padding="VALID", w_init=w_init)(x)
        x = nn.relu(x)
        x = hk.Conv2D(64, kernel_shape=3, stride=1, padding="VALID", w_init=w_init)(x)
        x = nn.relu(x)
        # Flatten the feature map.
        return hk.Flatten()(x)


class SACEncoder(hk.Module):
    """
    Encoder for SAC+AE.
    """

    def __init__(self, num_layers=4, num_filters=32, negative_slope=0.1):
        super().__init__()
        self.num_layers = num_layers
        self.num_filters = num_filters
        self.negative_slope = negative_slope

    def __call__(self, x):
        # Floatify the image.
        x = x.astype(jnp.float32) / 255.0

        # Apply CNN.
        w_init = DeltaOrthogonal(scale=np.sqrt(2 / (1 + self.negative_slope ** 2)))
        x = hk.Conv2D(self.num_filters, kernel_shape=4, stride=2, padding="VALID", w_init=w_init)(x)
        x = nn.leaky_relu(x, self.negative_slope)
        for _ in range(self.num_layers - 1):
            x = hk.Conv2D(self.num_filters, kernel_shape=3, stride=1, padding="VALID", w_init=w_init)(x)
            x = nn.leaky_relu(x, self.negative_slope)
        # Flatten the feature map.
        return hk.Flatten()(x)


class SACDecoder(hk.Module):
    """
    Decoder for SAC+AE.
    """

    def __init__(self, state_space, num_layers=4, num_filters=32, negative_slope=0.1):
        super().__init__()
        self.state_space = state_space
        self.num_layers = num_layers
        self.num_filters = num_filters
        self.negative_slope = negative_slope
        self.map_size = 43 - 2 * num_layers
        self.last_conv_dim = num_filters * self.map_size * self.map_size

    def __call__(self, x):
        # Apply linear layer.
        w_init = hk.initializers.Orthogonal(scale=np.sqrt(2 / (1 + self.negative_slope ** 2)))
        x = hk.Linear(self.last_conv_dim, w_init=w_init)(x)
        x = nn.leaky_relu(x, self.negative_slope).reshape(-1, self.map_size, self.map_size, self.num_filters)

        # Apply Transposed CNN.
        w_init = DeltaOrthogonal(scale=np.sqrt(2 / (1 + self.negative_slope ** 2)))
        for _ in range(self.num_layers - 1):
            x = hk.Conv2DTranspose(self.num_filters, kernel_shape=3, stride=1, padding="VALID", w_init=w_init)(x)
            x = nn.leaky_relu(x, self.negative_slope)

        # Apply output layer.
        w_init = DeltaOrthogonal(scale=1.0)
        x = hk.Conv2DTranspose(self.state_space.shape[2], kernel_shape=4, stride=2, padding="VALID", w_init=w_init)(x)
        return x


class SLACEncoder(hk.Module):
    """
    Encoder for SLAC.
    """

    def __init__(self, output_dim=256, negative_slope=0.2):
        super().__init__()
        self.output_dim = output_dim
        self.negative_slope = negative_slope

    def __call__(self, x):
        B, S, H, W, C = x.shape

        # Floatify the image.
        x = x.astype(jnp.float32) / 255.0
        # Reshape.
        x = x.reshape([B * S, H, W, C])
        # Apply CNN.
        w_init = DeltaOrthogonal(scale=1.0)
        depth = [32, 64, 128, 256, self.output_dim]
        kernel = [5, 3, 3, 3, 4]
        stride = [2, 2, 2, 2, 1]
        padding = ["SAME", "SAME", "SAME", "SAME", "VALID"]

        for i in range(5):
            x = hk.Conv2D(
                depth[i],
                kernel_shape=kernel[i],
                stride=stride[i],
                padding=padding[i],
                w_init=w_init,
            )(x)
            x = nn.leaky_relu(x, self.negative_slope)

        return x.reshape([B, S, -1])


class SLACDecoder(hk.Module):
    """
    Decoder for SLAC.
    """

    def __init__(self, state_space, std=1.0, negative_slope=0.2):
        super().__init__()
        self.state_space = state_space
        self.std = std
        self.negative_slope = negative_slope

    def __call__(self, x):
        B, S, latent_dim = x.shape

        # Reshape.
        x = x.reshape([B * S, 1, 1, latent_dim])

        # Apply CNN.
        w_init = DeltaOrthogonal(scale=1.0)
        depth = [256, 128, 64, 32, self.state_space.shape[2]]
        kernel = [4, 3, 3, 3, 5]
        stride = [1, 2, 2, 2, 2]
        padding = ["VALID", "SAME", "SAME", "SAME", "SAME"]

        for i in range(4):
            x = hk.Conv2DTranspose(
                depth[i],
                kernel_shape=kernel[i],
                stride=stride[i],
                padding=padding[i],
                w_init=w_init,
            )(x)
            x = nn.leaky_relu(x, self.negative_slope)

        x = hk.Conv2DTranspose(
            depth[-1],
            kernel_shape=kernel[-1],
            stride=stride[-1],
            padding=padding[-1],
            w_init=w_init,
        )(x)

        _, W, H, C = x.shape
        x = x.reshape([B, S, W, H, C])
        return x, jax.lax.stop_gradient(jnp.ones_like(x) * self.std)



from collections import deque

import numpy as np
from gym.spaces import Box, Discrete


class NStepBuffer:
    """
    Buffer for calculating n-step returns.
    """

    def __init__(
        self,
        gamma=0.99,
        nstep=3,
    ):
        self.discount = [gamma ** i for i in range(nstep)]
        self.nstep = nstep
        self.state = deque(maxlen=self.nstep)
        self.action = deque(maxlen=self.nstep)
        self.reward = deque(maxlen=self.nstep)

    def append(self, state, action, reward):
        self.state.append(state)
        self.action.append(action)
        self.reward.append(reward)

    def get(self):
        assert len(self.reward) > 0

        state = self.state.popleft()
        action = self.action.popleft()
        reward = self.nstep_reward()
        return state, action, reward

    def nstep_reward(self):
        reward = np.sum([r * d for r, d in zip(self.reward, self.discount)])
        self.reward.popleft()
        return reward

    def is_empty(self):
        return len(self.reward) == 0

    def is_full(self):
        return len(self.reward) == self.nstep

    def __len__(self):
        return len(self.reward)


class ReplayBuffer:
    """
    Replay Buffer.
    """

    def __init__(
        self,
        buffer_size,
        state_space,
        action_space,
        gamma,
        nstep,
    ):
        assert len(state_space.shape) in (1, 3)

        self._n = 0
        self._p = 0
        self.buffer_size = buffer_size
        self.nstep = nstep
        self.state_shape = state_space.shape
        self.use_image = len(self.state_shape) == 3

        if self.use_image:
            # Store images as a list of LazyFrames, which uses 4 times less memory.
            self.state = [None] * buffer_size
            self.next_state = [None] * buffer_size
        else:
            self.state = np.empty((buffer_size, *state_space.shape), dtype=np.float32)
            self.next_state = np.empty((buffer_size, *state_space.shape), dtype=np.float32)

        if type(action_space) == Box:
            self.action = np.empty((buffer_size, *action_space.shape), dtype=np.float32)
        elif type(action_space) == Discrete:
            self.action = np.empty((buffer_size, 1), dtype=np.int32)
        else:
            NotImplementedError

        self.reward = np.empty((buffer_size, 1), dtype=np.float32)
        self.done = np.empty((buffer_size, 1), dtype=np.float32)

        if nstep != 1:
            self.nstep_buffer = NStepBuffer(gamma, nstep)

    def append(self, state, action, reward, done, next_state, episode_done=None):

        if self.nstep != 1:
            self.nstep_buffer.append(state, action, reward)

            if self.nstep_buffer.is_full():
                state, action, reward = self.nstep_buffer.get()
                self._append(state, action, reward, done, next_state)

            if done or episode_done:
                while not self.nstep_buffer.is_empty():
                    state, action, reward = self.nstep_buffer.get()
                    self._append(state, action, reward, done, next_state)

        else:
            self._append(state, action, reward, done, next_state)

    def _append(self, state, action, reward, done, next_state):
        self.state[self._p] = state
        self.action[self._p] = action
        self.reward[self._p] = float(reward)
        self.done[self._p] = float(done)
        self.next_state[self._p] = next_state

        self._p = (self._p + 1) % self.buffer_size
        self._n = min(self._n + 1, self.buffer_size)

    def _sample_idx(self, batch_size):
        return np.random.randint(low=0, high=self._n, size=batch_size)

    def _sample(self, idxes):
        if self.use_image:
            state = np.empty((len(idxes), *self.state_shape), dtype=np.uint8)
            next_state = state.copy()
            for i, idx in enumerate(idxes):
                state[i, ...] = self.state[idx]
                next_state[i, ...] = self.next_state[idx]
        else:
            state = self.state[idxes]
            next_state = self.next_state[idxes]

        return (
            state,
            self.action[idxes],
            self.reward[idxes],
            self.done[idxes],
            next_state,
        )

    def sample(self, batch_size):
        idxes = self._sample_idx(batch_size)
        batch = self._sample(idxes)
        # Use fake weight to use the same interface with PER.
        weight = np.ones((), dtype=np.float32)
        return weight, batch



from functools import partial
from typing import Tuple

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
from jax.experimental import optix

from rljax.algorithm.base_class import OffPolicyActorCritic
from rljax.network import ContinuousQFunction, DeterministicPolicy
from rljax.util import add_noise, optimize


class DDPG(OffPolicyActorCritic):
    name = "DDPG"

    def __init__(
        self,
        num_agent_steps,
        state_space,
        action_space,
        seed,
        max_grad_norm=None,
        gamma=0.99,
        nstep=1,
        num_critics=1,
        buffer_size=10 ** 6,
        use_per=False,
        batch_size=256,
        start_steps=10000,
        update_interval=1,
        tau=5e-3,
        fn_actor=None,
        fn_critic=None,
        lr_actor=1e-3,
        lr_critic=1e-3,
        units_actor=(256, 256),
        units_critic=(256, 256),
        d2rl=False,
        std=0.1,
        update_interval_policy=2,
    ):
        super(DDPG, self).__init__(
            num_agent_steps=num_agent_steps,
            state_space=state_space,
            action_space=action_space,
            seed=seed,
            max_grad_norm=max_grad_norm,
            gamma=gamma,
            nstep=nstep,
            num_critics=num_critics,
            buffer_size=buffer_size,
            use_per=use_per,
            batch_size=batch_size,
            start_steps=start_steps,
            update_interval=update_interval,
            tau=tau,
        )
        if d2rl:
            self.name += "-D2RL"

        if fn_critic is None:

            def fn_critic(s, a):
                return ContinuousQFunction(
                    num_critics=num_critics,
                    hidden_units=units_critic,
                    d2rl=d2rl,
                )(s, a)

        if fn_actor is None:

            def fn_actor(s):
                return DeterministicPolicy(
                    action_space=action_space,
                    hidden_units=units_actor,
                    d2rl=d2rl,
                )(s)

        # Critic.
        self.critic = hk.without_apply_rng(hk.transform(fn_critic))
        self.params_critic = self.params_critic_target = self.critic.init(next(self.rng), *self.fake_args_critic)
        opt_init, self.opt_critic = optix.adam(lr_critic)
        self.opt_state_critic = opt_init(self.params_critic)

        # Actor.
        self.actor = hk.without_apply_rng(hk.transform(fn_actor))
        self.params_actor = self.params_actor_target = self.actor.init(next(self.rng), *self.fake_args_actor)
        opt_init, self.opt_actor = optix.adam(lr_actor)
        self.opt_state_actor = opt_init(self.params_actor)

        # Other parameters.
        self.std = std
        self.update_interval_policy = update_interval_policy

    @partial(jax.jit, static_argnums=0)
    def _select_action(
        self,
        params_actor: hk.Params,
        state: np.ndarray,
    ) -> jnp.ndarray:
        return self.actor.apply(params_actor, state)

    @partial(jax.jit, static_argnums=0)
    def _explore(
        self,
        params_actor: hk.Params,
        state: np.ndarray,
        key: jnp.ndarray,
    ) -> jnp.ndarray:
        action = self.actor.apply(params_actor, state)
        return add_noise(action, key, self.std, -1.0, 1.0)

    def update(self, writer=None):
        self.learning_step += 1
        weight, batch = self.buffer.sample(self.batch_size)
        state, action, reward, done, next_state = batch

        # Update critic and target.
        self.opt_state_critic, self.params_critic, loss_critic, abs_td = optimize(
            self._loss_critic,
            self.opt_critic,
            self.opt_state_critic,
            self.params_critic,
            self.max_grad_norm,
            params_actor_target=self.params_actor_target,
            params_critic_target=self.params_critic_target,
            state=state,
            action=action,
            reward=reward,
            done=done,
            next_state=next_state,
            weight=weight,
            **self.kwargs_critic,
        )
        self.params_critic_target = self._update_target(self.params_critic_target, self.params_critic)

        # Update priority.
        if self.use_per:
            self.buffer.update_priority(abs_td)

        # Update actor and target.
        if self.learning_step % self.update_interval_policy == 0:
            self.opt_state_actor, self.params_actor, loss_actor, _ = optimize(
                self._loss_actor,
                self.opt_actor,
                self.opt_state_actor,
                self.params_actor,
                self.max_grad_norm,
                params_critic=self.params_critic,
                state=state,
                **self.kwargs_actor,
            )
            self.params_actor_target = self._update_target(self.params_actor_target, self.params_actor)

            if writer and self.learning_step % 1000 == 0:
                writer.add_scalar("loss/critic", loss_critic, self.learning_step)
                writer.add_scalar("loss/actor", loss_actor, self.learning_step)

    @partial(jax.jit, static_argnums=0)
    def _sample_action(
        self,
        params_actor: hk.Params,
        state: np.ndarray,
    ) -> jnp.ndarray:
        return self.actor.apply(params_actor, state)

    @partial(jax.jit, static_argnums=0)
    def _calculate_target(
        self,
        params_critic_target: hk.Params,
        reward: np.ndarray,
        done: np.ndarray,
        next_state: np.ndarray,
        next_action: jnp.ndarray,
    ) -> jnp.ndarray:
        next_q = self._calculate_value(params_critic_target, next_state, next_action)
        return jax.lax.stop_gradient(reward + (1.0 - done) * self.discount * next_q)

    @partial(jax.jit, static_argnums=0)
    def _loss_critic(
        self,
        params_critic: hk.Params,
        params_critic_target: hk.Params,
        params_actor_target: hk.Params,
        state: np.ndarray,
        action: np.ndarray,
        reward: np.ndarray,
        done: np.ndarray,
        next_state: np.ndarray,
        weight: np.ndarray,
        *args,
        **kwargs,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        next_action = self._sample_action(params_actor_target, next_state, *args, **kwargs)
        target = self._calculate_target(params_critic_target, reward, done, next_state, next_action)
        q_list = self._calculate_value_list(params_critic, state, action)
        return self._calculate_loss_critic_and_abs_td(q_list, target, weight)

    @partial(jax.jit, static_argnums=0)
    def _loss_actor(
        self,
        params_actor: hk.Params,
        params_critic: hk.Params,
        state: np.ndarray,
    ) -> jnp.ndarray:
        action = self.actor.apply(params_actor, state)
        mean_q = self.critic.apply(params_critic, state, action)[0].mean()
        return -mean_q, None



from functools import partial
from typing import Tuple

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np

from rljax.algorithm.ddpg import DDPG
from rljax.util import add_noise


class TD3(DDPG):
    name = "TD3"

    def __init__(
        self,
        num_agent_steps,
        state_space,
        action_space,
        seed,
        max_grad_norm=None,
        gamma=0.99,
        nstep=1,
        num_critics=2,
        buffer_size=10 ** 6,
        use_per=False,
        batch_size=256,
        start_steps=10000,
        update_interval=1,
        tau=5e-3,
        fn_actor=None,
        fn_critic=None,
        lr_actor=1e-3,
        lr_critic=1e-3,
        units_actor=(256, 256),
        units_critic=(256, 256),
        d2rl=False,
        std=0.1,
        std_target=0.2,
        clip_noise=0.5,
        update_interval_policy=2,
    ):
        if not hasattr(self, "use_key_critic"):
            self.use_key_critic = True

        super(TD3, self).__init__(
            num_agent_steps=num_agent_steps,
            state_space=state_space,
            action_space=action_space,
            seed=seed,
            max_grad_norm=max_grad_norm,
            gamma=gamma,
            nstep=nstep,
            num_critics=num_critics,
            buffer_size=buffer_size,
            use_per=use_per,
            batch_size=batch_size,
            start_steps=start_steps,
            update_interval=update_interval,
            tau=tau,
            fn_actor=fn_actor,
            fn_critic=fn_critic,
            std=std,
            update_interval_policy=update_interval_policy,
        )
        self.std_target = std_target
        self.clip_noise = clip_noise

    @partial(jax.jit, static_argnums=0)
    def _sample_action(
        self,
        params_actor: hk.Params,
        state: np.ndarray,
        key: jnp.ndarray,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        action = self.actor.apply(params_actor, state)
        return add_noise(action, key, self.std_target, -1.0, 1.0, -self.clip_noise, self.clip_noise)



import gym
import numpy as np
from gym.spaces import Box


def make_continuous_env(env_id):
    return NormalizedActionEnv(gym.make(env_id))


class NormalizedActionEnv(gym.Wrapper):
    def __init__(self, env):
        gym.Wrapper.__init__(self, env)
        self._max_episode_steps = env._max_episode_steps
        # Original action space.
        self._low = env.action_space.low
        self._delta = env.action_space.high - env.action_space.low
        # Normalized action space.
        self.action_space = Box(low=-1.0, high=1.0, shape=env.action_space.shape, dtype=np.float64)

    def step(self, action):
        action = self._convert_action(action)
        return self.env.step(action)

    def _convert_action(self, action):
        return (action + 1.0) * self._delta / 2.0 + self._low



from typing import Sequence

import haiku as hk
import jax
import jax.numpy as jnp


class DeltaOrthogonal(hk.initializers.Initializer):
    """
    Delta-orthogonal initializer.
    """

    def __init__(self, scale=1.0, axis=-1):
        self.scale = scale
        self.axis = axis

    def __call__(self, shape: Sequence[int], dtype) -> jnp.ndarray:
        if len(shape) not in [3, 4, 5]:
            raise ValueError("Delta orthogonal initializer requires 3D, 4D or 5D shape.")
        w_mat = jnp.zeros(shape, dtype=dtype)
        w_orthogonal = hk.initializers.Orthogonal(self.scale, self.axis)(shape[-2:], dtype)
        if len(shape) == 3:
            k = shape[0]
            return jax.ops.index_update(
                w_mat,
                jax.ops.index[(k - 1) // 2, ...],
                w_orthogonal,
            )
        elif len(shape) == 4:
            k1, k2 = shape[:2]
            return jax.ops.index_update(
                w_mat,
                jax.ops.index[(k1 - 1) // 2, (k2 - 1) // 2, ...],
                w_orthogonal,
            )
        else:
            k1, k2, k3 = shape[:3]
            return jax.ops.index_update(
                w_mat,
                jax.ops.index[(k1 - 1) // 2, (k2 - 1) // 2, (k3 - 1) // 2, ...],
                w_orthogonal,
            )



from functools import partial
from typing import List, Tuple

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np

from rljax.algorithm.sac import SAC
from rljax.network import CategoricalPolicy, DiscreteQFunction
from rljax.util import get_q_at_action


class SAC_Discrete(SAC):
    name = "SAC-Discrete"

    def __init__(
        self,
        num_agent_steps,
        state_space,
        action_space,
        seed,
        max_grad_norm=None,
        gamma=0.99,
        nstep=1,
        num_critics=2,
        buffer_size=10 ** 6,
        use_per=False,
        batch_size=64,
        start_steps=10000,
        update_interval=4,
        update_interval_target=8000,
        fn_actor=None,
        fn_critic=None,
        lr_actor=3e-4,
        lr_critic=3e-4,
        lr_alpha=3e-4,
        units_actor=(512,),
        units_critic=(512,),
        target_entropy_ratio=0.98,
        dueling_net=False,
    ):
        if fn_critic is None:

            def fn_critic(s):
                return DiscreteQFunction(
                    action_space=action_space,
                    num_critics=num_critics,
                    hidden_units=units_critic,
                    dueling_net=dueling_net,
                )(s)

        if fn_actor is None:

            def fn_actor(s):
                return CategoricalPolicy(
                    action_space=action_space,
                    hidden_units=units_actor,
                )(s)

        # Entropy coefficient.
        self.target_entropy = -np.log(1.0 / action_space.n) * target_entropy_ratio

        if not hasattr(self, "use_key_critic"):
            self.use_key_critic = False
        if not hasattr(self, "use_key_actor"):
            self.use_key_actor = False

        super(SAC_Discrete, self).__init__(
            num_agent_steps=num_agent_steps,
            state_space=state_space,
            action_space=action_space,
            seed=seed,
            max_grad_norm=max_grad_norm,
            gamma=gamma,
            nstep=nstep,
            num_critics=num_critics,
            buffer_size=buffer_size,
            use_per=use_per,
            batch_size=batch_size,
            start_steps=start_steps,
            update_interval=update_interval,
            update_interval_target=update_interval_target,
            fn_actor=fn_actor,
            fn_critic=fn_critic,
            lr_actor=lr_actor,
            lr_critic=lr_critic,
            lr_alpha=lr_alpha,
        )

    @partial(jax.jit, static_argnums=0)
    def _select_action(
        self,
        params_actor: hk.Params,
        state: np.ndarray,
    ) -> jnp.ndarray:
        pi_s, _ = self.actor.apply(params_actor, state)
        return jnp.argmax(pi_s, axis=1)

    @partial(jax.jit, static_argnums=0)
    def _explore(
        self,
        params_actor: hk.Params,
        state: np.ndarray,
        key: jnp.ndarray,
    ) -> jnp.ndarray:
        pi_s, _ = self.actor.apply(params_actor, state)
        return jax.random.categorical(key, pi_s)

    @partial(jax.jit, static_argnums=0)
    def _sample_action(
        self,
        params_actor: hk.Params,
        state: np.ndarray,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        return self.actor.apply(params_actor, state)

    @partial(jax.jit, static_argnums=0)
    def _calculate_value_list(
        self,
        params_critic: hk.Params,
        state: np.ndarray,
        action: np.ndarray,
    ) -> List[jnp.ndarray]:
        return [get_q_at_action(q_s, action) for q_s in self.critic.apply(params_critic, state)]

    @partial(jax.jit, static_argnums=0)
    def _calculate_value(
        self,
        params_critic: hk.Params,
        state: np.ndarray,
        pi_s: np.ndarray,
    ) -> jnp.ndarray:
        q_s = jax.lax.stop_gradient(jnp.asarray(self.critic.apply(params_critic, state)).min(axis=0))
        return (pi_s * q_s).sum(axis=1, keepdims=True)

    @partial(jax.jit, static_argnums=0)
    def _calculate_log_pi(
        self,
        pi_s: np.ndarray,
        log_pi_s: np.ndarray,
    ) -> jnp.ndarray:
        return (pi_s * log_pi_s).sum(axis=1, keepdims=True)



from collections import deque

import numpy as np
from gym.spaces import Box, Discrete


class LazyFrames:
    """
    Stacked frames which never allocate memory to the same frame.
    """

    def __init__(self, frames):
        self._frames = list(frames)

    def __array__(self, dtype=None):
        out = np.array(self._frames)
        if out is not None:
            out = out.astype(dtype)
        return out

    def __len__(self):
        return len(self._frames)


class SequenceBuffer:
    """
    Buffer for storing sequence data.
    """

    def __init__(self, num_sequences=8):
        self.num_sequences = num_sequences
        self._reset_episode = False
        self.state_ = deque(maxlen=self.num_sequences + 1)
        self.action_ = deque(maxlen=self.num_sequences)
        self.reward_ = deque(maxlen=self.num_sequences)
        self.done_ = deque(maxlen=self.num_sequences)

    def reset(self):
        self._reset_episode = False
        self.state_.clear()
        self.action_.clear()
        self.reward_.clear()
        self.done_.clear()

    def reset_episode(self, state):
        assert not self._reset_episode
        self._reset_episode = True
        self.state_.append(state)

    def append(self, action, reward, done, next_state):
        assert self._reset_episode
        self.action_.append(action)
        self.reward_.append(reward)
        self.done_.append(done)
        self.state_.append(next_state)

    def get(self):
        state_ = LazyFrames(self.state_)
        action_ = np.array(self.action_)
        reward_ = np.array(self.reward_)
        done_ = np.array(self.done_)
        return state_, action_, reward_, done_

    def is_empty(self):
        return len(self.reward_) == 0

    def is_full(self):
        return len(self.reward_) == self.num_sequences

    def __len__(self):
        return len(self.reward_)


class SLACReplayBuffer:
    """
    Replay Buffer for SLAC.
    """

    def __init__(
        self,
        buffer_size,
        state_space,
        action_space,
        num_sequences,
    ):
        assert len(state_space.shape) in (1, 3)

        self._n = 0
        self._p = 0
        self.buffer_size = buffer_size
        self.num_sequences = num_sequences
        self.state_shape = state_space.shape
        self.use_image = len(self.state_shape) == 3

        if self.use_image:
            # Store images as a list of LazyFrames, which uses 4 times less memory.
            self.state_ = [None] * buffer_size
        else:
            self.state_ = np.empty((buffer_size, num_sequences + 1, *state_space.shape), dtype=np.float32)

        if type(action_space) == Box:
            self.action_ = np.empty((buffer_size, num_sequences, *action_space.shape), dtype=np.float32)
        elif type(action_space) == Discrete:
            self.action_ = np.empty((buffer_size, num_sequences, 1), dtype=np.int32)
        else:
            NotImplementedError

        self.reward_ = np.empty((buffer_size, num_sequences, 1), dtype=np.float32)
        self.done_ = np.empty((buffer_size, num_sequences, 1), dtype=np.float32)

        # Buffer to store a sequence of trajectories.
        self.seq_buffer = SequenceBuffer(num_sequences)

    def reset_episode(self, state):
        """
        Reset the sequence buffer and set the initial observation. This has to be done before every episode starts.
        """
        self.seq_buffer.reset_episode(state)

    def append(self, action, reward, done, next_state, episode_done=None):
        self.seq_buffer.append(action, reward, done, next_state)

        if self.seq_buffer.is_full():
            state_, action_, reward_, done_ = self.seq_buffer.get()
            self._append(state_, action_, reward_, done_)

        if episode_done:
            self.seq_buffer.reset()

    def _append(self, state_, action_, reward_, done_):
        self.state_[self._p] = state_
        self.action_[self._p] = action_.reshape(self.num_sequences, -1)
        self.reward_[self._p] = reward_.reshape(self.num_sequences, 1)
        self.done_[self._p] = done_.reshape(self.num_sequences, 1)

        self._p = (self._p + 1) % self.buffer_size
        self._n = min(self._n + 1, self.buffer_size)

    def _sample_idx(self, batch_size):
        return np.random.randint(low=0, high=self._n, size=batch_size)

    def _sample_state(self, idxes):
        if self.use_image:
            state_ = np.empty((len(idxes), self.num_sequences + 1, *self.state_shape), dtype=np.uint8)
            for i, idx in enumerate(idxes):
                state_[i, ...] = self.state_[idx]
        else:
            state_ = self.state_[idxes]
        return state_

    def sample_model(self, batch_size):
        idxes = self._sample_idx(batch_size)
        return (self._sample_state(idxes), self.action_[idxes], self.reward_[idxes], self.done_[idxes])

    def sample_sac(self, batch_size):
        idxes = self._sample_idx(batch_size)
        return (self._sample_state(idxes), self.action_[idxes], self.reward_[idxes, -1], self.done_[idxes, -1])



import os
from datetime import timedelta
from time import sleep, time

import pandas as pd
from tensorboardX import SummaryWriter


class Trainer:
    """
    Trainer.
    """

    def __init__(
        self,
        env,
        env_test,
        algo,
        log_dir,
        seed=0,
        action_repeat=1,
        num_agent_steps=10 ** 6,
        eval_interval=10 ** 4,
        num_eval_episodes=10,
        save_params=False,
    ):
        assert num_agent_steps % action_repeat == 0
        assert eval_interval % action_repeat == 0

        # Envs.
        self.env = env
        self.env_test = env_test

        # Set seeds.
        self.env.seed(seed)
        self.env_test.seed(2 ** 31 - seed)

        # Algorithm.
        self.algo = algo

        # Log setting.
        self.log = {"step": [], "return": []}
        self.csv_path = os.path.join(log_dir, "log.csv")
        self.param_dir = os.path.join(log_dir, "param")
        self.writer = SummaryWriter(log_dir=os.path.join(log_dir, "summary"))

        # Other parameters.
        self.action_repeat = action_repeat
        self.num_agent_steps = num_agent_steps
        self.eval_interval = eval_interval
        self.num_eval_episodes = num_eval_episodes
        self.save_params = save_params

    def train(self):
        # Time to start training.
        self.start_time = time()
        # Initialize the environment.
        state = self.env.reset()

        for step in range(1, self.num_agent_steps + 1):
            state = self.algo.step(self.env, state)

            if self.algo.is_update():
                self.algo.update(self.writer)

            if step % self.eval_interval == 0:
                self.evaluate(step)
                if self.save_params:
                    self.algo.save_params(os.path.join(self.param_dir, f"step{step}"))

        # Wait for the logging to be finished.
        sleep(2)

    def evaluate(self, step):
        total_return = 0.0
        for _ in range(self.num_eval_episodes):
            state = self.env_test.reset()
            done = False
            while not done:
                action = self.algo.select_action(state)
                state, reward, done, _ = self.env_test.step(action)
                total_return += reward

        # Log mean return.
        mean_return = total_return / self.num_eval_episodes
        # To TensorBoard.
        self.writer.add_scalar("return/test", mean_return, step * self.action_repeat)
        # To CSV.
        self.log["step"].append(step * self.action_repeat)
        self.log["return"].append(mean_return)
        pd.DataFrame(self.log).to_csv(self.csv_path, index=False)

        # Log to standard output.
        print(f"Num steps: {step * self.action_repeat:<6}   Return: {mean_return:<5.1f}   Time: {self.time}")

    @property
    def time(self):
        return str(timedelta(seconds=int(time() - self.start_time)))
